/*
 * adc.c part of XMEGA Oscilloscope
 *
 * Created: 2016-12-16 09:40:21
 * Atmel Studio 6 (Version: 6.2.1563 - Service Pack 2)
 *  Author: Krzysztof Markiewicz
 *  www.obbo.pl
 *
 * This program is distributed under the terms of the GNU General Public License
 */ 

#include <avr/interrupt.h>
#include <util/delay.h>
#include "adc.h"
#include "mso.h"
#include "mathasm.h"
#include "bitlib.h"


void adc_DMAReset(void);
void adc_DMASetTriggerSource(uint8_t chcfg, uint8_t samplingrate);
void adc_DMAStopChannel(DMA_CH_t *dma);
void adc_DMASetChannel(DMA_CH_t *dma, uint16_t bufferlength, volatile uint8_t *source, uint8_t *dest);
uint8_t adc_CalibrationTune(uint8_t val, const int8_t step);


void adc_Init(void) {
    ADCA.CALL = ReadCalibrationByte(offsetof(NVM_PROD_SIGNATURES_t, ADCACAL0) );
    ADCB.CALL = ReadCalibrationByte(offsetof(NVM_PROD_SIGNATURES_t, ADCBCAL0) );
    ADCA.CALH = ReadCalibrationByte(offsetof(NVM_PROD_SIGNATURES_t, ADCACAL1) );
    ADCB.CALH = ReadCalibrationByte(offsetof(NVM_PROD_SIGNATURES_t, ADCBCAL1) );
    ADCA.CTRLA = ADC_ENABLE_bm;				
    ADCB.CTRLA = ADC_ENABLE_bm;				
    ADCA.REFCTRL = ADC_REFSEL_AREFA_gc;		// REF = (2.5V / 2)
    ADCB.REFCTRL = ADC_REFSEL_AREFA_gc;		// REF = (2.5V / 2)
    ADCA.CH0.MUXCTRL = ADC_CH_MUXPOS_PIN5_gc | ADC_CH_MUXNEG_PIN6_gc;
    ADCB.CH0.MUXCTRL = ADC_CH_MUXPOS_PIN5_gc | ADC_CH_MUXNEG_PIN6_gc;
}

void adc_InitResolution8bit(void) {
	ADCA.CTRLB = ADC_IMPMODE_bm | ADC_CONMODE_bm | ADC_RESOLUTION_8BIT_gc;
	ADCB.CTRLB = ADC_IMPMODE_bm | ADC_CONMODE_bm | ADC_RESOLUTION_8BIT_gc;
}

void adc_InitResolution12bit(void) {
	ADCA.CTRLB = ADC_IMPMODE_bm | ADC_CONMODE_bm | ADC_RESOLUTION_12BIT_gc;
	ADCB.CTRLB = ADC_IMPMODE_bm | ADC_CONMODE_bm | ADC_RESOLUTION_12BIT_gc;
}

void adc_InitGain(uint8_t ch1gain, uint8_t ch2gain) {
	ADCA.CH0.CTRL = ADC_CH_INPUTMODE_DIFFWGAIN_gc | (input_ChannelADCGain(ch1gain) << 2);
	ADCB.CH0.CTRL = ADC_CH_INPUTMODE_DIFFWGAIN_gc | (input_ChannelADCGain(ch2gain) << 2);
}

void adc_InitTrigered(void) {
	ADCA.PRESCALER = ADC_PRESCALER_DIV512_gc;
	ADCB.PRESCALER = ADC_PRESCALER_DIV512_gc;
}

void adc_BufferStop(void) {
	adc_DMAStop();
}

void adc_DMAReset(void) {
	DMA.CH0.CTRLA |= DMA_CH_RESET_bm;
	DMA.CH1.CTRLA |= DMA_CH_RESET_bm;
	DMA.CH2.CTRLA |= DMA_CH_RESET_bm;
}

void adc_DMAStart(uint8_t chcfg) {
	ADCA.CTRLA |= ADC_FLUSH_bm;
	ADCB.CTRLA |= ADC_FLUSH_bm;
	if (testbit(chcfg, MAIN_CHANNEL_CH1)) {
		DMA.CH0.CTRLA |= DMA_CH_ENABLE_bm;
	}
	if (testbit(chcfg, MAIN_CHANNEL_CH2)) {	
		DMA.CH1.CTRLA |= DMA_CH_ENABLE_bm;
	}
	if (testbit(chcfg, MAIN_CHANNEL_CHD)) {
		DMA.CH2.CTRLA |= DMA_CH_ENABLE_bm;
	}
}

void adc_DMAStopChannel(DMA_CH_t *dma) {
	uint8_t t = 100;
	
	dma->CTRLA &= ~(DMA_CH_ENABLE_bm);
	while ((dma->CTRLA & DMA_CH_ENABLE_bm) && (t != 0)) {
		_delay_us(1);
		t--;
	}
	dma->TRIGSRC = DMA_CH_TRIGSRC_OFF_gc;
}

void adc_DMAStop(void) {
	adc_DMAStopChannel(&DMA.CH0);
	adc_DMAStopChannel(&DMA.CH1);
	adc_DMAStopChannel(&DMA.CH2);
}

void adc_DMASetTriggerSource(uint8_t chcfg, uint8_t samplingrate) {
	uint8_t source;
	
	// for sampling rate < 6 set one trigger source for all channel, ADC trigger source should be tie with active DMA channel
	if (samplingrate < SAMPLING_RATE_32ksps) {
		if (testbit(chcfg, MAIN_CHANNEL_CH1)) {
			source = DMA_CH_TRIGSRC_ADCA_CH0_gc;
		} else if (testbit(chcfg, MAIN_CHANNEL_CH2)) {
			source = DMA_CH_TRIGSRC_ADCB_CH0_gc;
		} else {
			source = DMA_CH_TRIGSRC_EVSYS_CH0_gc;
		}
		DMA.CH0.TRIGSRC = source;
		DMA.CH1.TRIGSRC = source;
		DMA.CH2.TRIGSRC = source;
	} else {
		if (testbit(chcfg, MAIN_CHANNEL_CH1)) {
			DMA.CH0.TRIGSRC = DMA_CH_TRIGSRC_ADCA_CH0_gc;
		} 
		if (testbit(chcfg, MAIN_CHANNEL_CH2)) {
			DMA.CH1.TRIGSRC = DMA_CH_TRIGSRC_ADCB_CH0_gc;
		} 
		if (testbit(chcfg, MAIN_CHANNEL_CHD)) {
			DMA.CH2.TRIGSRC = DMA_CH_TRIGSRC_EVSYS_CH0_gc;
		}
	}
}

void adc_DMASetChannel(DMA_CH_t *dma, uint16_t bufferlength, volatile uint8_t *source, uint8_t *dest) {
	dma->ADDRCTRL  = DMA_CH_DESTRELOAD_BLOCK_gc | DMA_CH_DESTDIR_INC_gc;
	dma->TRFCNT    = bufferlength;
	dma->SRCADDR0  = ((uint16_t)(source)) & 0xff;
	dma->SRCADDR1  = (((uint16_t)(source)) >> 8) & 0xff;
	dma->SRCADDR2  = 0;
	dma->DESTADDR0 = ((uint16_t)(dest)) & 0xff;
	dma->DESTADDR1 = (((uint16_t)(dest)) >> 8) & 0xff;
	dma->DESTADDR2 = 0;
	dma->CTRLA     = DMA_CH_REPEAT_bm | DMA_CH_SINGLE_bm | DMA_CH_BURSTLEN_1BYTE_gc;
}

void adc_DMASet(uint16_t bufferlength, uint8_t chcfg, databuffer_t buffer, uint8_t samplingrate) {
	adc_DMAReset();
	// DMA.CH0 for ADC CH1
	if (testbit(chcfg, MAIN_CHANNEL_CH1)) {
		adc_DMASetChannel(&DMA.CH0, bufferlength, &ADCA.CH0.RESL, buffer.ch0);
	}
	// DMA.CH1 for ADC CH2
	if (testbit(chcfg, MAIN_CHANNEL_CH2)) {
		adc_DMASetChannel(&DMA.CH1, bufferlength, &ADCB.CH0.RESL, buffer.ch1);
	}
	// DMA.CH2 for Logic
	if (testbit(chcfg, MAIN_CHANNEL_CHD)) {
		adc_DMASetChannel(&DMA.CH2, bufferlength, &VPORT2.IN, buffer.chd);
	}
	adc_DMASetTriggerSource(chcfg, samplingrate);
	adc_DMAStart(chcfg);
}

int8_t adc_ReadChannel8bit(ADC_t *channel) {
	return subsat(0, channel->CH0.RESL);
}

