// Minimal USB Stack for ATxmega32a4u and related
// http://nonolithlabs.com
// (C) 2011 Kevin Mehall (Nonolith Labs) <km@kevinmehall.net>
//
// Heavily borrows from LUFA
// Copyright 2011  Dean Camera (dean [at] fourwalledcubicle [dot] com)
//
// Licensed under the terms of the GNU GPLv3+


#ifndef USB_XMEGA_H_
#define USB_XMEGA_H_

#include "data.h"
#include "bitlib.h"


#pragma once

// export definition to Xlab oscyloskopx.h
#define USB_SEND_CONFIGURATION		'A'		// 65  0x41
#define USB_SEND_CONFIGURATION_ACK	'B'		// 66  0x42
#define USB_SEND_CALIBRATE			'C'		// 67  0x43
#define USB_SEND_CONTROL			'D'		// 68  0x44
#define USB_SEND_DATA_BULK			'E'		// 69  0x45
#define USB_SEND_FREQUENCY			'F'		// 70  0x46
#define USB_REQUEST_DATA			'G'		// 71  0x47
#define USB_DISCONNECT				'H'		// 72  0x48
// end export definition to Xlab oscyloskopx.h

#define CPU_TO_LE16(x) x

#include "StdDescriptors.h"
#include "StdRequestType.h"

#define USB_MAXEP				1			// Maximum endpoint address
#define USB_EP0SIZE				64			// Endpoint zero size
#define USB_EP1SIZE				64

#define USB_TIMEOUT_10us		20000       

typedef struct USB_EP_pair {
	USB_EP_t out;
	USB_EP_t in;
} ATTR_PACKED USB_EP_pair_t;

extern USB_EP_pair_t endpoints[USB_MAXEP + 1];	// Endpoints configuration and status table

/** String descriptor index for the device's unique serial number string descriptor within the device.
 *  This unique serial number is used by the host to associate resources to the device (such as drivers or COM port
 *  number allocations) to a device regardless of the port it is plugged in to on the host. Some microcontrollers contain
 *  a unique serial number internally, and setting the device descriptors serial number string index to this value
 *  will cause it to use the internal serial number.
 *
 *  On unsupported devices, this will evaluate to \ref NO_DESCRIPTOR and so will force the host to create a pseudo-serial
 *  number for the device. */
#define USE_INTERNAL_SERIAL            NO_DESCRIPTOR

/** Length of the device's unique internal serial number, in bits, if present on the selected microcontroller model. */
#define INTERNAL_SERIAL_LENGTH_BITS    (8 * (1 + (offsetof(NVM_PROD_SIGNATURES_t, COORDY1) - offsetof(NVM_PROD_SIGNATURES_t, LOTNUM0))))

/** Start address of the internal serial number, in the appropriate address space, if present on the selected microcontroller
 *  model. */
#define INTERNAL_SERIAL_START_ADDRESS  offsetof(NVM_PROD_SIGNATURES_t, LOTNUM0)

/** Enum for the various states of the USB Device state machine. Only some states are
	*  implemented in the LUFA library - other states are left to the user to implement.
	*
	*  For information on each possible USB device state, refer to the USB 2.0 specification.
	*
	*  \see \ref USB_DeviceState, which stores the current device state machine state. */
enum USB_Device_States_t {
	DEVICE_STATE_Unattached           = 0, /**< Internally implemented by the library. This state indicates that the
											*   device is not currently connected to a host. */
	DEVICE_STATE_Powered              = 1, /**< Internally implemented by the library. This state indicates that the
											*   device is connected to a host, but enumeration has not yet begun.  */
	DEVICE_STATE_Default              = 2, /**< Internally implemented by the library. This state indicates that the
											*   device's USB bus has been reset by the host and it is now waiting for
											*   the host to begin the enumeration process. */
	DEVICE_STATE_Addressed            = 3, /**< Internally implemented by the library. This state indicates that the
											*   device has been addressed by the USB Host, but is not yet begun */
	DEVICE_STATE_Configured           = 4, /**< May be implemented by the user project. This state indicates that the
											*   device has been enumerated by the host and is ready for USB
											*   communications to begin.  */
	DEVICE_STATE_Suspended            = 5, /**< May be implemented by the user project. This state indicates that the
											*   USB bus has been suspended by the host, and the device should
											*   power down to a minimal power level until the bus is resumed. */
};

#define USB_EP_size_to_gc(x)  ((x <= 8   )?USB_EP_BUFSIZE_8_gc:\
                               (x <= 16  )?USB_EP_BUFSIZE_16_gc:\
                               (x <= 32  )?USB_EP_BUFSIZE_32_gc:\
                               (x <= 64  )?USB_EP_BUFSIZE_64_gc:\
                               (x <= 128 )?USB_EP_BUFSIZE_128_gc:\
                               (x <= 256 )?USB_EP_BUFSIZE_256_gc:\
                               (x <= 512 )?USB_EP_BUFSIZE_512_gc:\
                                           USB_EP_BUFSIZE_1023_gc)

extern volatile uint8_t USB_DeviceState;
extern volatile uint8_t USB_Device_ConfigurationNumber;

void USB_EndpointsConfiguration(void);
void USB_InterfaceAttach(void);	// Initialize USB module and endpoints
void USB_InterfaceDetach(void);

void USB_ep0_send_progmem(const uint8_t* addr, uint8_t size);

uint16_t CALLBACK_USB_GetDescriptor(const uint16_t wValue,
                                    const uint8_t wIndex,
                                    const void** const DescriptorAddress);



#endif /* USB_XMEGA_H_ */