/*
 * main.c part of XMEGA Oscilloscope
 *
 * Created: 2016-12-16
 * Atmel Studio 6 (Version: 6.2.1563 - Service Pack 2)
 *  Author: Krzysztof Markiewicz
 *  www.obbo.pl
 *
 * This program is distributed under the terms of the GNU General Public License
 */ 


// Hardware resources:
/*  Timers:
        TCC0 Source is Event EVSYS.CH7 (39,94ms)
			 Meter - Frequency counter time keeper, 
			 MSO - Controls the auto trigger
        TCC1 Meter - Frequency, pulse, period counter low 16bits, source is CPUclk or EVSYS.CH2
			 MSO - Counts post trigger samples, source is Event EVSYS.CH1
        TCD0 Split timer, source is Event EVSYS.CH6 (1.024ms)
             MSO - TCD0L 39,94ms period - 25,04 Hz source for EVSYS.CH7
	         AWG - TCD0H Controls sweep refresh
	    TCD1 AWG - Overflow used for EVSYS.CH3
        TCE0 MSO - Controls Interrupt ADC (sampling rate >= 11) and strobe ADC for sampling rate: 6, 7, 8, 9, 10   
             Meter - Frequency, pulse, period counter high 16bits
		TCE1 Meter - duty counter low 16bits
		TCF0 Meter - duty counter high 16bits, 
			 Sniffer - Timeout
    Events:
	    CH0 TCE0 overflow used for start conversion ADC and strobe DMA for logic channel
	    CH1 ADC.CH0 conversion complete for post trigger counter
        CH2 Sniffer - strobe DMA,
		    Meter - Transfer input pin for frequency, pulse, period measuring
        CH3 TCD1 overflow used for DAC
        CH4 TCC0 overflow used for frequency, pulse, period measuring
        CH5 TCC1 overflow used for frequency, pulse, period measuring
        CH6 CPUclk / 32768 for TCD0
        CH7 TCD0L underflow: 39,94ms period - 25,04 Hz for TCC0
	DMAs:
	    CH0 ADC CH0  / Sniffer SPI.MOSI / Sniffer USART Rx
	    CH1 ADC CH1  / Sniffer SPI.MISO / Sniffer USART Tx
	    CH2 PortC CHD / Sniffer SPI, USART
	    CH3 DAC
    USART:
	    USARTC0 for sniffer USART Rx
		USARTC1 for sniffer USART Tx
	SPI:
	    SPIC for sniffer SPI.MOSI
	    SPIE for sniffer SPI.MISO
    Interrupt Levels:
        PORTC INT0      High        I2C sniffer
        TCE0            High        Slow Sampling
        USARTC0 RXC     High	    USART sniffer
		USARTC1 RXC     High	    USART sniffer
        USB BUSEVENT    Medium      USB Bus Event
        USB_TRNCOMPL    Medium      USB Transaction Complete
        TCC0L           Low         auto trigger
		OSC				NMI			XOSC or PLL failure
*/

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/fuse.h>
#include <stdbool.h>
#include "main.h"
#include "awg.h"
#include "dac.h"
#include "sniffer.h"
#include "mso.h"
#include "multimeter.h"
#include "usb.h"
#include "USB\usb_xmega.h"

FUSES = {
	.FUSEBYTE0 = 0x00,
    .FUSEBYTE1 = 0x00,  /* Watchdog closed window periods = 8CLK, Watchdog timeout periods = 8CLK */
    .FUSEBYTE2 = 0xfd,  /* Reset vector = Application reset, TOSC1/2 shared with XTAL, BOD enabled in sampled mode */
    .FUSEBYTE4 = 0xf7,  /* STARTUPTIME = 4ms, Watchdog timer not locked, JTAG disabled */
    .FUSEBYTE5 = 0xe9,  /* BOD enabled continuously, EEPROM is erased during chip erase, BODLVL = 2V8 */
};

/*
	JTAGUSERID = 0x00
	WDWP = 8CLK
	WDP = 8CLK
	BOOTRST = APPLICATION
	TOSCSEL = XTAL
	BODPD = SAMPLED
	RSTDISBL = [ ]
	SUT = 4MS
	WDLOCK = [ ]
	JTAGEN = [ ]
	BODACT = CONTINUOUS
	EESAVE = [ ]
	BODLVL = 2V8
*/

#define PORTE_PIN7_TEST_CLOCK_ENABLE	0	// to enable 2Mhz clock signal on PORTE, PORTE.PIN7 must be prior disconnected from PORTC.PIN7

uint8_t EEMEM eecalibrated = 0;				// ADC offset calibration done

void CCPWrite( volatile uint8_t *address, uint8_t value );
void main_ClearUSBError(void);
void main_GetMSOPreBuffer(uint8_t sizech, uint8_t chcfg, databuffer_t buffer, uint8_t samplingrate, bool logicdelay);
void main_StopDMASampling(uint8_t samplingrate);
void main_MeasureVAC(const uint16_t length, uint8_t chctrl);
void main_WaitForConnection(void);
void main_ShowStatus(uint8_t errors, uint8_t status);
void main_IOPortsInit(void);
void main_PowerReduction(void);


int main(void) {
	// USB Clock
	DFLLRC32M.CALB = ReadCalibrationByte(offsetof(NVM_PROD_SIGNATURES_t, USBRCOSC));		
	DFLLRC32M.COMP1 = 0x1b;		// Xmega AU manual, p100                                     
	DFLLRC32M.COMP2 = 0xb7;                  
	OSC.DFLLCTRL |= OSC_RC32MCREF_USBSOF_gc;												
	DFLLRC32M.CTRL |= DFLL_ENABLE_bm;														
	// System clock
    OSC.XOSCCTRL |= OSC_FRQRANGE_12TO16_gc | OSC_XOSCSEL_XTAL_16KCLK_gc;
    OSC.PLLCTRL |= OSC_PLLSRC_XOSC_gc | OSC_PLLFAC1_bm;
    OSC.CTRL |= OSC_RC32MEN_bm | OSC_XOSCEN_bm;
    _delay_ms(2);               
    OSC.CTRL |= OSC_PLLEN_bm;
	// Wait for stable oscillator
	setbit(GPIO_Errors, MAIN_ERROR_MainClock);
	for (uint8_t i = 0; i < 200; i++) {
		if (OSC.STATUS & OSC_XOSCRDY_bm) {
			if (OSC.STATUS & OSC_PLLRDY_bm) {              
				CCPWrite(&CLK.CTRL, CLK_SCLKSEL_PLL_gc);				
				clrbit(GPIO_Errors, MAIN_ERROR_MainClock);													
				CCPWrite(&OSC.XOSCFAIL, (OSC_PLLFDEN_bm | OSC_XOSCFDEN_bm));
				break;
			}
		}
		_delay_us(10);
	}
	// Ports configuration
	main_IOPortsInit();
	// Test wave on PortE Pin7
	#if PORTE_PIN7_TEST_CLOCK_ENABLE == 1
		VPORT1.DIR = PIN7_bm;
		PORTCFG.CLKEVOUT = PORTCFG_EVOUT_PE7_gc;
		EVSYS.CH0MUX = EVSYS_CHMUX_PRESCALER_16_gc;
	#endif
	// Event System
	EVSYS.CH3MUX = EVSYS_CHMUX_TCD1_OVF_gc;			// Used for DAC
    EVSYS.CH6MUX = EVSYS_CHMUX_PRESCALER_32768_gc;  // CLKPER / 32768 -> 1.024ms
    EVSYS.CH7MUX = EVSYS_CHMUX_TCD0_OVF_gc;
	// Timers TCD0
	TCD0.CTRLA = TC_CLKSEL_EVCH6_gc;
	TCD0.CTRLE = TC_BYTEM_SPLITMODE_gc;
	TCD0.PERL = 39;									// 39,94ms auto trigger base
	TCD0.PERH = 100;								// sweep refresh
    // DAC
	dac_Init();
	dac_On();
    // ADC
	adc_Init();
	// DMA
    DMA.CTRL = DMA_ENABLE_bm;							
    // Interrupt configuration
    PMIC.CTRL = PMIC_HILVLEN_bm | PMIC_MEDLVLEN_bm | PMIC_LOLVLEN_bm;
    sei();
	// Power reduction
	main_PowerReduction();
	for (;;) {		
		uint8_t result, dacctrl, chctrl, ch1ctrl, ch2ctrl, chdctrl, prev_mainmode, prev_samplingrate, triggertype;	
		
		// Check status
		if (eeprom_read_byte(&eecalibrated)) setbit(GPIO_Status, MAIN_STATUS_Calibrated);
		while (1) {
			main_ShowStatus(GPIO_Errors, GPIO_Status);
			if (!GPIO_Errors) break;
			main_ClearUSBError();
			_delay_ms(500);
		}
		data_LoadDefaultConfiguration();
		data_MainBufferInit( master.dacbufferblocks, master.chbufferblocks, 1, data_ChannelEnabled(1, 1, 1));
		USB_InterfaceAttach();	
		// Wait for USB connection
		while (!testbit(GPIO_Status, MAIN_STATUS_USB_Connected)) {
			main_WaitForConnection();
		}
        // Application loop
		usbbuffers.config->id = 0;
        while ((testbit(GPIO_Status, MAIN_STATUS_USB_Connected)) && (!GPIO_Errors)) {
			clrbit(GPIO_Request, MAIN_REQUEST_Update);
			clrbit(GPIO_Action, MSO_TRIGGER_STATUS_Break);
			triggertype = GPIO_Trigger;
			prev_samplingrate = GPIO_SampleRate;
			GPIO_SampleRate = master.samplerate;
			prev_mainmode = running.mainmode;
			running = master;
			ch1ctrl = GPIO_CH1ctrl;
			ch2ctrl = GPIO_CH2ctrl;
			chdctrl = GPIO_CHDctrl;
			dacctrl = GPIO_DACctrl;
			chctrl = data_ChannelEnabled(ch1ctrl, ch2ctrl, chdctrl);
			// Clear scope
			if ((running.mainmode != prev_mainmode) && ((prev_mainmode == MAIN_MODE_Scope) || (prev_mainmode == MAIN_MODE_MeterVAC))) {
				mso_Stop();
				clrbit(GPIO_Request, MAIN_REQUEST_USB_Send);
			}
			if ((prev_samplingrate != GPIO_SampleRate) && (GPIO_SampleRate >= SAMPLING_RATE_800sps)) {
				if (testbit(ch1ctrl, MAIN_CHANNEL_CONTROL_Enable)) data_BufferClear8(usbbuffers.ch1, running.chbufferblocks * DATA_BUFFER_BLOCK_SIZE, 0);
				if (testbit(ch2ctrl, MAIN_CHANNEL_CONTROL_Enable)) data_BufferClear8(usbbuffers.ch2, running.chbufferblocks * DATA_BUFFER_BLOCK_SIZE, 0);
				if (testbit(chdctrl, MAIN_CHANNEL_CONTROL_Enable)) data_BufferClear8(usbbuffers.chd, running.chbufferblocks * DATA_BUFFER_BLOCK_SIZE, 0);
			} 
			if ((running.mainmode == MAIN_MODE_Scope) || (running.mainmode == MAIN_MODE_MeterVAC)) {
				if ((GPIO_SampleRate >= SAMPLING_RATE_800sps) && (prev_samplingrate < SAMPLING_RATE_800sps)) {
					mso_TimersStop();
					mso_DMASamplingStop();
				}
				if ((GPIO_SampleRate < SAMPLING_RATE_800sps) && (prev_samplingrate >= SAMPLING_RATE_800sps)) {
					mso_SlowSamplingStop();
					clrbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered);
				}
				adc_DMAStop();
			}
			// BUFFER control
			if (testbit(GPIO_Request, MAIN_REQUEST_BufferResize)) {
				clrbit(GPIO_Request, MAIN_REQUEST_BufferResize);
				running.dacbufferblocks = master.dacbufferblocks;
				running.chbufferblocks = master.chbufferblocks;
				dac_BufferStop();
				adc_BufferStop();
				data_MainBufferInit(running.dacbufferblocks, running.chbufferblocks, dacctrl, chctrl);
				data_BufferClear16(awgbuffers.dac, running.dacbufferblocks * DATA_BUFFER_BLOCK_SIZE, DAC_OUT_Zero);
				data_BufferClear8(usbbuffers.all, data_FrameLength(running.chbufferblocks, chctrl), 0);
				mm_MeasureClearResult(usbbuffers.mmch1);
				mm_MeasureClearResult(usbbuffers.mmch2);
				usbbuffers.mmch1->offset = data_ReadCHDataOffset(MAIN_CHANNEL_CH1, ch1ctrl, running.ch1gain);
				usbbuffers.mmch2->offset = data_ReadCHDataOffset(MAIN_CHANNEL_CH2, ch2ctrl, running.ch2gain);
				setbit(GPIO_Request, MAIN_REQUEST_AWG_Update);
				setbit(GPIO_Request, MAIN_REQUEST_AWG_BufferUpdate);
				clrbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered);
				setbit(GPIO_Action, MSO_TRIGGER_STATUS_Break);
				setbit(GPIO_Request, MAIN_REQUEST_Update);
			}
			// Clear multimeter results
			if (testbit(GPIO_Request, MAIN_REQUEST_MM_Clear)) {
				clrbit(GPIO_Request, MAIN_REQUEST_MM_Clear);
				mm_MeasureClearResult(usbbuffers.mmch1);
				mm_MeasureClearResult(usbbuffers.mmch2);
			}
			// Input configuration
			if (testbit(GPIO_Request, MAIN_REQUEST_MSO_Update)) {
				uint8_t clearch = 0;
				clrbit(GPIO_Request, MAIN_REQUEST_MSO_Update);  
				if ((usbbuffers.mmch1->gain != master.ch1gain) && (GPIO_SampleRate >= SAMPLING_RATE_800sps) && (testbit(ch1ctrl, MAIN_CHANNEL_CONTROL_Enable))) {
					setbit(clearch, MAIN_CHANNEL_CH1);
				}
				running.ch1gain = master.ch1gain;
				if ((usbbuffers.mmch2->gain != master.ch2gain) && (GPIO_SampleRate >= SAMPLING_RATE_800sps) && (testbit(ch2ctrl, MAIN_CHANNEL_CONTROL_Enable))) {
					setbit(clearch, MAIN_CHANNEL_CH2);
				}
				running.ch2gain = master.ch2gain;
				if ((running.mainmode == MAIN_MODE_Scope) || (running.mainmode == MAIN_MODE_MeterVAC) || (running.mainmode == MAIN_MODE_MeterVDC)) {
					input_AnalogPortConfig(ch1ctrl, running.ch1gain, ch2ctrl, running.ch2gain);
					usbbuffers.mmch1->offset = data_ReadCHDataOffset(MAIN_CHANNEL_CH1, ch1ctrl, running.ch1gain);
					usbbuffers.mmch1->control = ch1ctrl;
					usbbuffers.mmch2->offset = data_ReadCHDataOffset(MAIN_CHANNEL_CH2, ch2ctrl, running.ch2gain);
					usbbuffers.mmch2->control = ch2ctrl;
					usbbuffers.mmchd->control = chdctrl;
					usbbuffers.mmchd->bitenable = GPIO_CHDmask;
					adc_InitGain(running.ch1gain, running.ch2gain);
					data_StoreCHDataGain(running.ch1gain, running.ch2gain);
					_delay_ms(50);
					if (testbit(clearch, MAIN_CHANNEL_CH1)) data_BufferClear8(usbbuffers.ch1, running.chbufferblocks * DATA_BUFFER_BLOCK_SIZE, 0);
					if (testbit(clearch, MAIN_CHANNEL_CH2)) data_BufferClear8(usbbuffers.ch2, running.chbufferblocks * DATA_BUFFER_BLOCK_SIZE, 0);
				}
				input_LogicPortConfig(&CHD_PORT, testbit(chdctrl, MAIN_CHANNEL_CONTROL_Invert), INPUT_PORT_SENSE_EDGE_Rising);
			}
		    // MULTIMETER, LOGIC SNIFFER, SCOPE control
			if ((running.mainmode == MAIN_MODE_MeterVDC) && (chctrl)) {
				mm_MeasureVDC(GPIO_CH1ctrl, running.ch1gain, usbbuffers.mmch1, GPIO_CH2ctrl, running.ch2gain, usbbuffers.mmch2);
			} else if (running.mainmode ==  MAIN_MODE_MeterFrequency) {
				mm_MeasureFrequency(&master, GPIO_TriggerSource, usbbuffers.mmch1, testbit(GPIO_MeterFrequency, METER_FREQUENCY_FastMode));
			} else if (running.mainmode ==  MAIN_MODE_MeterPeriod) {
				mm_MeasurePeriod(&master, GPIO_TriggerSource, usbbuffers.mmch1, usbbuffers.mmch2);
			} else if (running.mainmode ==  MAIN_MODE_MeterCount) {
				mm_CountPulses(&master, GPIO_TriggerSource, usbbuffers.mmch1);
			} else if (running.mainmode ==  MAIN_MODE_MeterGate) {
				mm_CountPulsesWithGate(&master, GPIO_TriggerSource, GPIO_GateSource, usbbuffers.mmch1, GPIO_MeterFrequency);
			} else if ((running.mainmode == MAIN_MODE_Sniff) && (testbit(GPIO_Action, SNIFFER_REQUEST_Run)))	{
			    result = sniffer_Decode(&master.mainmode, sniffbuffers, &running, GPIO_SniffOption, GPIO_Mixed);
				if (result) setbit(GPIO_Errors, MAIN_ERROR_USB);
			} else if (running.mainmode == MAIN_MODE_MeterVAC) {
				// Start acquiring data
				usbbuffers.mmch1->samplingrate = GPIO_SampleRate;
				usbbuffers.mmch2->samplingrate = GPIO_SampleRate;
				if (!(testbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered))) {
					mso_ConfigApply(&running, GPIO_SampleRate);
					if (GPIO_SampleRate < SAMPLING_RATE_800sps) {
						main_GetMSOPreBuffer(running.chbufferblocks, chctrl, databuffers, GPIO_SampleRate, false);
					} else {
						mso_SlowSamplingStart();
					}
					setbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered);
				}
				// Finish acquiring data (triggered)
				if (GPIO_SampleRate < SAMPLING_RATE_800sps) {
					clrbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered);
					main_StopDMASampling(GPIO_SampleRate);
					usbbuffers.config->index = mso_DMARestToTransfer(chctrl);
					mso_BufferDateCalculate(data_ChannelDataLength(running.chbufferblocks) + ADC_BUFFER_EXTEND - usbbuffers.config->index,
											running.chbufferblocks * DATA_BUFFER_BLOCK_SIZE, 
											chctrl, ch1ctrl, ch2ctrl, chdctrl,
											databuffers, usbbuffers);
				} else {	
					if (testbit(GPIO_Status, MAIN_STATUS_Acquired)) {
						clrbit(GPIO_Status, MAIN_STATUS_Acquired);
						clrbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered);
					}
				}
				main_MeasureVAC(running.chbufferblocks * DATA_BUFFER_BLOCK_SIZE, chctrl);
			} else if (running.mainmode == MAIN_MODE_Scope) {
				// Start acquiring data
				usbbuffers.mmch1->samplingrate = GPIO_SampleRate;
				usbbuffers.mmch2->samplingrate = GPIO_SampleRate;
				if (!(testbit(GPIO_Action, MSO_TRIGGER_STATUS_Stop)) && !(testbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered))) {
					mso_ConfigApply(&running, GPIO_SampleRate);
					if (testbit(triggertype, MAIN_TRIGGER_Free)) {
						if ((GPIO_SampleRate < SAMPLING_RATE_800sps) || ((GPIO_SampleRate >= SAMPLING_RATE_800sps) && (!testbit(GPIO_Mixed, MAIN_MSO_MODE_Roll)))) {
							mso_TriggerHold(running.triggerhold);
						}
					}
					if (GPIO_SampleRate < SAMPLING_RATE_800sps) {
						main_GetMSOPreBuffer(running.chbufferblocks, chctrl, databuffers, GPIO_SampleRate, true);
					}
					if (testbit(triggertype, MAIN_TRIGGER_Free) || (testbit(GPIO_Mixed, MAIN_MSO_MODE_Roll) && (GPIO_SampleRate >= SAMPLING_RATE_800sps))) {
						setbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered);
					} else {
						mso_WaitForTriger(&running, chctrl, triggertype, GPIO_TriggerSource, GPIO_SampleRate, usbbuffers.mmch1, usbbuffers.mmch2, ch1ctrl, ch2ctrl, chdctrl);
					}
					if (testbit(GPIO_Action, MSO_TRIGGER_STATUS_Force)) {
						clrbit(GPIO_Action, MSO_TRIGGER_STATUS_Force);
						setbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered);
					}
					if (GPIO_SampleRate >= SAMPLING_RATE_800sps) mso_SlowSamplingStart();
				}
				mso_DisableAutoTrigger();
				// Finish acquiring data (triggered)
				if (testbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered)) {
					if (GPIO_SampleRate < SAMPLING_RATE_800sps) {
						clrbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered);
						main_StopDMASampling(GPIO_SampleRate);
						usbbuffers.config->index = mso_DMARestToTransfer(chctrl);
						mso_BufferDateCalculate(data_ChannelDataLength(running.chbufferblocks) + ADC_BUFFER_EXTEND - usbbuffers.config->index,
												running.chbufferblocks * DATA_BUFFER_BLOCK_SIZE,
												chctrl, ch1ctrl, ch2ctrl, chdctrl,
												databuffers, usbbuffers);
						setbit(GPIO_Request, MAIN_REQUEST_USB_Send);
					} else {
						if (testbit(GPIO_Status, MAIN_STATUS_Acquired)) {
							clrbit(GPIO_Status, MAIN_STATUS_Acquired);
							clrbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered);
						}
					}
					// MULTIMETER Voltage AC control
					if (testbit(GPIO_Mixed, MAIN_METER_AlwaysOn)) {
						main_MeasureVAC(running.chbufferblocks * DATA_BUFFER_BLOCK_SIZE, chctrl);
					}
					if (!testbit(GPIO_Mixed, MAIN_METER_AlwaysOn)) {
						setbit(GPIO_Request, MAIN_REQUEST_MM_Clear);
					}
					// Single trigger
					if (testbit(triggertype, MAIN_TRIGGER_Single)) {
						setbit(GPIO_Action, MSO_TRIGGER_STATUS_Stop);
						clrbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered);
					}
				}
				// USB send scope data
				if  (testbit(GPIO_Request, MAIN_REQUEST_USB_Send)) {
					clrbit(GPIO_Request, MAIN_REQUEST_USB_Send);
					usbbuffers.config->type = FRAME_HEADER_NAME_Scope;
					usbbuffers.config->chenable = chctrl;
					usbbuffers.config->chtype = data_ChannelAnalog();
					usbbuffers.config->chlength = running.chbufferblocks * DATA_BUFFER_BLOCK_SIZE;
					usbbuffers.config->sapmlingrate = GPIO_SampleRate;
					usbbuffers.config->id += 1;
					result = usb_SendDataBuffer(usbbuffers.all, data_FrameLength(running.chbufferblocks, chctrl));
					if (result) setbit(GPIO_Errors, MAIN_ERROR_USB);
				}
			}
	        // DAC control
	        if (testbit(GPIO_Request, MAIN_REQUEST_DAC_Update)) {
		        clrbit(GPIO_Request, MAIN_REQUEST_DAC_Update);
		        if (testbit(dacctrl, MAIN_CHANNEL_CONTROL_Enable)) {
			        dac_On();
			        setbit(GPIO_Request, MAIN_REQUEST_AWG_Update);
			    } else {
			        dac_Off();
		        }
	        }
			// AWG control
			if (testbit(dacctrl, MAIN_CHANNEL_CONTROL_Enable)) {
				if (testbit(GPIO_Request, MAIN_REQUEST_AWG_Update)) {
					if (master.wave.type == AWG_WAVE_WhiteNoise) {
						dac_DMAStop();
						awg_WaveWhiteNoise(&master.wave, (uint8_t *)(intptr_t)GPIO_DACctrl);
					} else if (master.wave.type == AWG_WAVE_PinkNoise) {
						dac_DMAStop();
						awg_WavePinkNoise(&master.wave, (uint8_t *)(intptr_t)GPIO_DACctrl);
					} else if (master.wave.type == AWG_WAVE_Constant) {
						dac_DMAStop();
						awg_WaveConstant(master.wave.constvalue);
					} else {
						clrbit(GPIO_Request, MAIN_REQUEST_AWG_Update);
						awg_WaveTemplate(awgbuffers.wave, running.dacbufferblocks * DATA_BUFFER_BLOCK_SIZE, master.wave.type);
						dac_DMAStart(awgbuffers.dac, running.dacbufferblocks * DATA_BUFFER_BLOCK_SIZE);
						setbit(GPIO_Request, MAIN_REQUEST_AWG_BufferUpdate);
					}
				}
				if (awg_SweepEnabled(master.sweep.type)) {
					if (testbit(TCD2.INTFLAGS, TC2_HUNFIF_bp)) {
						setbit(TCD2.INTFLAGS, TC2_HUNFIF_bp);
						awg_Sweep(&master.sweep);
						setbit(GPIO_Request, MAIN_REQUEST_AWG_BufferUpdate);
					}
				}
				if ((testbit(GPIO_Request, MAIN_REQUEST_AWG_BufferUpdate)) && (!testbit(GPIO_Mixed, MAIN_AWG_UpdateDisable))) {
					clrbit(GPIO_Request, MAIN_REQUEST_AWG_BufferUpdate);
					awg_FillBuffer(awgbuffers, running.dacbufferblocks * DATA_BUFFER_BLOCK_SIZE, &master.sweep, &master.wave);
				}
			}
			// Application loop complete
	        LED_GRN_ON();
        }       
		// Stop module 
		dac_Off();
		mso_Stop();
		USB_InterfaceDetach();
		data_LoadDefaultConfiguration();
		input_AnalogPortConfig(GPIO_CH1ctrl, master.ch1gain, GPIO_CH2ctrl, master.ch2gain);
    }        
    return 0;
}

//ISR(osc_oscf_vect)
ISR(OSC_OSCF_vect) {
	setbit(GPIO_Errors, MAIN_ERROR_MainClock);										
    // Switch to internal 2MHz if crystal fails
    if (!testbit(OSC.STATUS, OSC_XOSCRDY_bp)) {		// XT ready flag not set?
        OSC.XOSCCTRL = 0x00;						// Disable external oscillators
        OSC.PLLCTRL = 0x10;							// 2MHz is PLL Source - 16x Factor (32MHz)
    }
}

// From Application Note AVR1003
void CCPWrite( volatile uint8_t *address, uint8_t value ) {
    uint8_t volatile saved_sreg = SREG;
    cli();

	#ifdef __ICCAVR__
		asm("movw r30, r16");
	#ifdef RAMPZ
		RAMPZ = 0;
	#endif
		asm("ldi  r16,  0xD8 \n"
			"out  0x34, r16  \n"
	#if (__MEMORY_MODEL__ == 1)
			"st     Z,  r17  \n");
	#elif (__MEMORY_MODEL__ == 2)
			"st     Z,  r18  \n");
	#else /* (__MEMORY_MODEL__ == 3) || (__MEMORY_MODEL__ == 5) */
			"st     Z,  r19  \n");
	#endif /* __MEMORY_MODEL__ */

	#elif defined __GNUC__
		volatile uint8_t *tmpAddr = address;
	#ifdef RAMPZ
		RAMPZ = 0;
	#endif
		asm volatile(
			"movw r30,  %0"	      "\n\t"
			"ldi  r16,  %2"	      "\n\t"
			"out   %3, r16"	      "\n\t"
			"st     Z,  %1"       "\n\t"
			:
			: "r" (tmpAddr), "r" (value), "M" (CCP_IOREG_gc), "i" (&CCP)
			: "r16", "r30", "r31"
			);

	#endif
		SREG = saved_sreg;
}

void main_GetMSOPreBuffer(uint8_t sizech, uint8_t chcfg, databuffer_t buffer, uint8_t samplingrate, bool logicdelay) {
	uint8_t result;
	
	adc_DMASet(data_ChannelDataLength(sizech), chcfg, buffer, samplingrate);
	if (logicdelay) mso_DMALogicChannelDelay(samplingrate, ADC_DMA_LOGIC_DELAY);
	mso_DMASamplingStart(samplingrate);
	if (chcfg) {
		result = mso_CaptureBufferPreTrigger(samplingrate, sizech);
		if (result) setbit(GPIO_Errors, MAIN_ERROR_MSO);
	}
}

void main_StopDMASampling(uint8_t samplingrate) {
	mso_WaitForSampleComplet(GPIO_SampleRate);
	mso_DMASamplingStop();
	mso_TimersStop();
	adc_DMAStop();
}

void main_MeasureVAC(const uint16_t length, uint8_t chctrl) {
	if (testbit(chctrl, MAIN_CHANNEL_CH1)) {
		mm_MeasureVAC((int8_t *)usbbuffers.ch1, length, usbbuffers.mmch1);
	} else {
		mm_MeasureClearResult(usbbuffers.mmch1);
	}
	if (testbit(chctrl, MAIN_CHANNEL_CH2)) {
		mm_MeasureVAC((int8_t *)usbbuffers.ch2, length, usbbuffers.mmch2);
	} else {
		mm_MeasureClearResult(usbbuffers.mmch2);
	}
}

void main_ClearUSBError(void) {
	clrbit(GPIO_Errors, MAIN_ERROR_USB);
	clrbit(GPIO_Errors, MAIN_ERROR_USBCRC);
}

void main_WaitForConnection(void) {
	LED_GRN_ON();
	_delay_ms(400);
	LED_GRN_OFF();
	_delay_ms(600);
}

void main_ShowStatus(uint8_t errors, uint8_t status) {
	uint8_t i;
	
	if (!testbit(status, MAIN_STATUS_Calibrated)) {
		LED_RED_ON();
	} else {
		LED_RED_OFF();
	}
	if (errors) {
		LED_RED_ON();
		_delay_ms(100);
		LED_RED_OFF();
		_delay_ms(100);
		LED_RED_ON();
		_delay_ms(100);
		LED_RED_OFF();
		_delay_ms(100);
		for (i = 0; i < 8; i++) {
			LED_GRN_ON();
			_delay_ms(100);
			LED_GRN_OFF();
			_delay_ms(100);
			if ((errors >> i) & 0x01) {
				LED_RED_ON();
			}
			_delay_ms(1000);
			LED_RED_OFF();
		}
	}
}

void main_IOPortsInit(void) {
    PORTCFG.VPCTRLA = PORTCFG_VP13MAP_PORTE_gc | PORTCFG_VP02MAP_PORTF_gc;		// VP1 Map to PORTE, VP0 Map to PORTF
    PORTCFG.VPCTRLB = PORTCFG_VP13MAP_PORTD_gc | PORTCFG_VP02MAP_PORTC_gc;		// VP3 Map to PORTD, VP2 Map to PORTC
	VPORT0.DIR = 0x00;															// PORTF as input
	VPORT1.DIR = 0x00;															// PORTE as input
	VPORT2.DIR = 0x00;															// PORTC as input
	VPORT3.DIR = PIN3_bm | PIN4_bm | PIN5_bm;									// Enable/Disable Logic Bus, LED_RED, LED_GREEN as output
	VPORT3.OUT &= ~(PIN3_bm);													// Enable Logic Bus
	PORTA.DIR = PIN3_bm | PIN4_bm;												// CH1 DIV10, AC/DC
	PORTB.DIR = PIN3_bm | PIN4_bm;												// CH2 DIV10, AC/DC
    PORTA.PIN5CTRL = PORT_ISC_INPUT_DISABLE_gc;									// VREF for ADC
    PORTA.PIN6CTRL = PORT_ISC_INPUT_DISABLE_gc;									// CH1 input
    PORTB.PIN5CTRL = PORT_ISC_INPUT_DISABLE_gc; 								// VREF for DAC
    PORTB.PIN6CTRL = PORT_ISC_INPUT_DISABLE_gc;									// CH2 input
	PORTD.PIN0CTRL = PORT_ISC_RISING_gc | PORT_INVEN_bm;						// EXT input
	PORTD.PIN1CTRL = PORT_ISC_RISING_gc;										// EXT input
}

void main_PowerReduction(void) {
    PR.PRGEN = PR_AES_bm | PR_EBI_bm | PR_RTC_bm;
    PR.PRPA = PR_DAC_bm | PR_AC_bm;
    PR.PRPB = PR_AC_bm;
	PR.PRPC = PR_TWI_bm | PR_USART0_bm | PR_USART1_bm | PR_SPI_bm;
    PR.PRPD = PR_TWI_bm | PR_USART0_bm | PR_USART1_bm | PR_SPI_bm;
    PR.PRPE = PR_TWI_bm | PR_USART0_bm | PR_USART1_bm | PR_SPI_bm;
    PR.PRPF = PR_TWI_bm | PR_USART0_bm | PR_USART1_bm | PR_SPI_bm;
}

