/*
 * mso.c part of XMEGA Oscilloscope
 *
 * Created: 2016-12-16 09:41:29
 * Atmel Studio 6 (Version: 6.2.1563 - Service Pack 2)
 *  Author: Krzysztof Markiewicz
 *  www.obbo.pl
 *
 * This program is distributed under the terms of the GNU General Public License
 */ 


#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include "mso.h"
#include "asmutil.h"
#include "input.h"
#include "mathasm.h"
#include "ADC.h"
#include "bitlib.h"



uint16_t mso_slowsamplecount;
volatile uint16_t mso_frameidx = 0;
// ADC system clock timers
// TCE0 controls Interrupt ADC, sampling rates: 6, 7, 8, 9, 10
const uint16_t mso_TCE0value[5] PROGMEM = {124, 249, 499, 1249, 2499};
// TCE0 timer is fixed for slow sampling rates > 10
// mso_counts controls the sampling rate.
const uint16_t mso_counts[11] PROGMEM = {2, 4, 8, 16, 40, 80, 160, 400, 800, 1600, 4000};


void mso_TimerPostTriggerInit(uint16_t posttrigger, uint8_t samplingrate, uint8_t source, uint8_t chcfg);
void mso_TimersInit(uint8_t timeout);
uint16_t mso_PostTrigerAdjust(uint16_t posttrigger, uint8_t samplingrate, uint8_t source);
void mso_TrigerCH1(config_t *cfg, uint8_t trigger, channela_t *channel, uint8_t chctrl);
void mso_TrigerCH2(config_t *cfg, uint8_t trigger, channela_t *channel, uint8_t chctrl);
int8_t mso_OffsetApply(int8_t chdata, int8_t offset);
int8_t mso_Invert(int8_t chdata);
void mso_MathApply(int8_t *ch1data, uint8_t ch1cfg, int8_t *ch2data, uint8_t ch2cfg);
int8_t mso_ElasticApply(int8_t prevdata, int8_t newdata);
void mso_EnableAutoTrigger(void);
uint8_t mso_CHDpreskaler(uint8_t samplingrate);
void mso_SlowSamplingControl(volatile uint16_t *index);
void mso_SamplingRateApply(uint8_t samplingrate, volatile uint16_t *index, uint16_t *samplecount);



/* Clock Timer table:
rate    S/s  Oversample ADCCLK  PRESCALE  TCE0   mso_counts        Method			 TCE0CLK
0         2M      x1     2 MHz     16     x          x	    DMA ADC free run 			x
1         1M      x1     1 MHz     32     x          x	    DMA ADC free run			x
2       500k      x1   500 kHz     64     x          x		DMA ADC free run			x
3       250k      x1   250 kHz    128     x          x	    DMA ADC free run			x
4       125k      x1   125 kHz    256     x          x		DMA ADC free run			x 
5      62.5k      x1  62.5 kHz    512     x          x	    DMA ADC free run			x
6        32k      x1   500 kHz     64     124        x          DMA w/Timer			   4MHz
7        16k      x1   500 kHz     64     249        x          DMA w/Timer			   4MHz
8         8k      x1   500 kHz     64     499        x          DMA w/Timer			   4MHz
9       3.2k      x1   500 kHz     64    1249        x          DMA w/Timer			   4MHz
10      1.6k      x1   500 kHz     64    2499        x          DMA w/Timer			   4MHz
11       800      x2   500 kHz     64    4999        2        ISR TCE0 1600Hz	 	   8MHz
12       320      x4   500 kHz     64    6249        4        ISR TCE0 1280Hz		   8MHz
13       160      x8   500 kHz     64    6249        8        ISR TCE0 1280Hz		   8MHz
14        80     x16   500 kHz     64    6249       16        ISR TCE0 1280Hz		   8MHz
15        32     x40   500 kHz     64    6249       40        ISR TCE0 1280Hz		   8MHz
16        16     x80   500 kHz     64    6249       80        ISR TCE0 1280Hz		   8MHz
17         8    x160   500 kHz     64    6249      160        ISR TCE0 1280Hz		   8MHz
18       3.2    x400   500 kHz     64    6249      400        ISR TCE0 1280Hz		   8MHz
19       1.6    x800   500 kHz     64    6249      800        ISR TCE0 1280Hz		   8MHz
20       0.8   x1600   500 kHz     64    6249     1600        ISR TCE0 1280Hz		   8MHz
21      0.32   x4000   500 kHz     64    6249     4000        ISR TCE0 1280Hz		   8MHz
*/



void mso_TrigerCH1(config_t *cfg, uint8_t trigger, channela_t *channel, uint8_t chctrl) {
	int8_t tlevel1, tlevel2;

    if (testbit(trigger, MAIN_TRIGGER_Window)) { 
		tlevel1 = subsat(channel->offset, cfg->window1);
		tlevel2 = subsat(channel->offset, cfg->window2);
		if (testbit(chctrl, MAIN_CHANNEL_CONTROL_Invert)) {
			windowCH1(-tlevel2, -tlevel1);
		} else {
			windowCH1(tlevel1, tlevel2);
		}
    } else if (testbit(trigger, MAIN_TRIGGER_Slope)) {
		tlevel1 = cfg->triggerlevel;
		if (tlevel1 < 1) tlevel1 = 1;
		if (testbit(chctrl, MAIN_CHANNEL_CONTROL_Invert)) {
			if (testbit(trigger, MAIN_TRIGGER_Falling)) slopeupCH1(tlevel1); else slopedownCH1(tlevel1);
		} else {
			if (testbit(trigger, MAIN_TRIGGER_Falling)) slopedownCH1(tlevel1); else slopeupCH1(tlevel1);
		}
    } else if (testbit(trigger, MAIN_TRIGGER_Edge)) {
		tlevel1 = subsat(channel->offset, cfg->triggerlevel);
		if (testbit(chctrl, MAIN_CHANNEL_CONTROL_Invert)) {
			if (testbit(trigger, MAIN_TRIGGER_Falling)) trigupCH1(-tlevel1); else trigdownCH1(-tlevel1);
		} else {
			if (testbit(trigger, MAIN_TRIGGER_Falling)) trigdownCH1(tlevel1); else trigupCH1(tlevel1);
		}
	} else {
		tlevel1 = subsat(channel->offset, cfg->triggerlevel);
		if (testbit(chctrl, MAIN_CHANNEL_CONTROL_Invert)) {
			if (ADCA.CH0.RESL < tlevel1) trigupCH1(-tlevel1); else trigdownCH1(-tlevel1); 
		} else {
			if (ADCA.CH0.RESL < tlevel1) trigdownCH1(tlevel1); else trigupCH1(tlevel1);
		}
	}
}

void mso_TrigerCH2(config_t *cfg, uint8_t trigger, channela_t *channel, uint8_t chctrl) {
	int8_t tlevel1, tlevel2;
    
    if (testbit(trigger, MAIN_TRIGGER_Window)) { 
		tlevel1 = subsat(channel->offset, cfg->window1);
		tlevel2 = subsat(channel->offset, cfg->window2);
		if (testbit(chctrl, MAIN_CHANNEL_CONTROL_Invert)) {
			windowCH2(-tlevel2, -tlevel1);
		} else {
			windowCH2(tlevel1, tlevel2);
		}
    } else if (testbit(trigger, MAIN_TRIGGER_Slope)) {
		tlevel1 = cfg->triggerlevel;
		if (tlevel1 < 1) tlevel1 = 1;
		if (testbit(chctrl, MAIN_CHANNEL_CONTROL_Invert)) {
			if (testbit(trigger, MAIN_TRIGGER_Falling)) slopeupCH2(tlevel1); else slopedownCH2(tlevel1);
		} else {
			if (testbit(trigger, MAIN_TRIGGER_Falling)) slopedownCH2(tlevel1); else slopeupCH2(tlevel1);
		}
    } else if (testbit(trigger, MAIN_TRIGGER_Edge)) {
		tlevel1 = subsat(channel->offset, cfg->triggerlevel);
		if (testbit(chctrl, MAIN_CHANNEL_CONTROL_Invert)) {
			if (testbit(trigger, MAIN_TRIGGER_Falling)) trigupCH2(-tlevel1); else trigdownCH2(-tlevel1);
		} else {
			if (testbit(trigger, MAIN_TRIGGER_Falling)) trigdownCH2(tlevel1); else trigupCH2(tlevel1);
		}
	} else {
		tlevel1 = subsat(channel->offset, cfg->triggerlevel);
		if (testbit(chctrl, MAIN_CHANNEL_CONTROL_Invert)) {
			if (ADCB.CH0.RESL < tlevel1) trigupCH2(-tlevel1); else trigdownCH2(-tlevel1);
		} else {
			if (ADCB.CH0.RESL < tlevel1) trigdownCH2(tlevel1); else trigupCH2(tlevel1);
		}
	}
}

void mso_WaitForTriger(config_t *cfg, uint8_t chcfg, uint8_t trigger, uint8_t source, uint8_t samplingrate, channela_t *channel1, channela_t *channel2, uint8_t ch1ctrl, uint8_t ch2ctrl, uint8_t chdctrl) {
	mso_TimerPostTriggerInit(cfg->triggerpos, samplingrate, source, chcfg);
	if (testbit(trigger, MAIN_TRIGGER_Auto)) mso_EnableAutoTrigger();
	if ((source == MAIN_TRIGGER_SOURCE_CH1) && testbit(ch1ctrl, MAIN_CHANNEL_CONTROL_Enable)) {
		mso_TrigerCH1(cfg, trigger, channel1, ch1ctrl);
	} else if ((source == MAIN_TRIGGER_SOURCE_CH2) && testbit(ch2ctrl, MAIN_CHANNEL_CONTROL_Enable)) {
		mso_TrigerCH2(cfg, trigger, channel2, ch2ctrl);
	} else if ((source <= TRIGGER_SOURCE_COUNT) && testbit(chdctrl, MAIN_CHANNEL_CONTROL_Enable)) {
		if (testbit(trigger, MAIN_TRIGGER_Falling)) {
			trigdownCHD(source - ADC_CHANNEL_ANALOG_COUNT);
		} else {
			trigupCHD(source - ADC_CHANNEL_ANALOG_COUNT);
		}
	}
	while (!testbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered) && !testbit(GPIO_Request, MAIN_REQUEST_Update)) {}
}

void mso_TimerPostTriggerInit(uint16_t posttrigger, uint8_t samplingrate, uint8_t source, uint8_t chcfg) {
	uint16_t tp;

	TCC1.CNT = 0;
	TCC1.INTFLAGS = TC1_OVFIF_bm;
	tp = mso_PostTrigerAdjust(posttrigger, samplingrate, source);
	TCC1.PERL = lobyte(tp);
	TCC1.PERH = hibyte(tp);
	if (testbit(chcfg, MAIN_CHANNEL_CH1)) {
		EVSYS.CH1MUX = EVSYS_CHMUX_ADCA_CH0_gc;
	} else if (testbit(chcfg, MAIN_CHANNEL_CH2)) {
		EVSYS.CH1MUX = EVSYS_CHMUX_ADCB_CH0_gc;
	} else {
		if (samplingrate < SAMPLING_RATE_32ksps) {
			EVSYS.CH1MUX = mso_CHDpreskaler(samplingrate);
		} else {
			EVSYS.CH1MUX = EVSYS_CHMUX_TCE0_OVF_gc;
		}
	}
}

uint16_t mso_PostTrigerAdjust(uint16_t posttrigger, uint8_t samplingrate, uint8_t source) {
	uint8_t adjust;
	
	// ATXmega-AU_Manual.pdf, ADC Clock and Conversion Timing p.346
	if (samplingrate == SAMPLING_RATE_2Msps) adjust = 6;
	else if (samplingrate <= SAMPLING_RATE_62k5ksps) adjust = 7;
	else adjust = 2;
	if ((source >= MAIN_TRIGGER_SOURCE_CHD_Bit0) && (source <= MAIN_TRIGGER_SOURCE_CHD_Bit7)) {
		if (samplingrate <= SAMPLING_RATE_62k5ksps) adjust = adjust - 5;
	}
	if (posttrigger > adjust) {
		return (posttrigger - adjust);
	} else {
		return 0;
	}
}

void mso_BufferDateCalculate(uint16_t index, 
							 uint16_t length, 
							 uint8_t chcfg, 
							 uint8_t ch1cfg, 
							 uint8_t ch2cfg, 
							 uint8_t chdcfg, 
							 databuffer_t datain, 
							 usbbuffer_t dataout) {   
    int8_t *in1, *in2, *ind;			// pointers to input (RAW data buffer)
    int8_t *out1, *out2, *outd;			// pointers to output (USB buffer)
    int8_t ch1, ch2;
    uint16_t i;

	if (index >= (length + ADC_BUFFER_EXTEND)) {
		index = index - (length + ADC_BUFFER_EXTEND);
	}
	in1 = (int8_t *)datain.ch0 + index;
	out1 = (int8_t *)dataout.ch1;
	in2 = (int8_t *)datain.ch1 + index;
	out2 = (int8_t *)dataout.ch2;
	ind = (int8_t *)datain.chd + index;
	outd = (int8_t *)dataout.chd;
	for (i = 0; i < length; i++) {
		if (testbit(chcfg, MAIN_CHANNEL_CH1)) {
			ch1 = mso_Invert(*in1++);
		}
	    if (testbit(chcfg, MAIN_CHANNEL_CH2)) {
			ch2 = mso_Invert(*in2++);
		}
		if (testbit(chcfg, MAIN_CHANNEL_CHD)) {
			*outd = *(ind++);
		}
		index++;
	    if (index >= length + ADC_BUFFER_EXTEND) {
		    index = 0;
		    in1 = (int8_t *)datain.ch0;
		    in2 = (int8_t *)datain.ch1;
			ind = (int8_t *)datain.chd;
	    }
		if (testbit(chcfg, MAIN_CHANNEL_CH1)) {
			ch1 = mso_OffsetApply(ch1, dataout.mmch1->offset);
			if (testbit(ch1cfg, MAIN_CHANNEL_CONTROL_Invert)) {
				ch1 = mso_Invert(ch1);
			}
		}
		if (testbit(chcfg, MAIN_CHANNEL_CH2)) {
			ch2 = mso_OffsetApply(ch2, dataout.mmch2->offset);
			if (testbit(ch2cfg, MAIN_CHANNEL_CONTROL_Invert)) {
				ch2 = mso_Invert(ch2);
			}
		}
		if ((testbit(chcfg, MAIN_CHANNEL_CH1)) && (testbit(chcfg, MAIN_CHANNEL_CH2))) {
			mso_MathApply(&ch1, ch1cfg, &ch2, ch2cfg);
		}
		if (testbit(chcfg, MAIN_CHANNEL_CH1)) {
			if (testbit(GPIO_Mixed, MAIN_MSO_MODE_Elastic)) {
				*out1 = mso_ElasticApply(*out1, ch1);
			} else {
				*out1 = ch1;
			}
			out1++; 
		}
		if (testbit(chcfg, MAIN_CHANNEL_CH2)) {
			if (testbit(GPIO_Mixed, MAIN_MSO_MODE_Elastic)) {
				*out2 = mso_ElasticApply(*out2, ch2);
			} else {
				*out2 = ch2;
			}
			out2++;
		}
		if (testbit(chcfg, MAIN_CHANNEL_CHD)) {
			outd++;
		}
    }
}

int8_t mso_OffsetApply(int8_t chdata, int8_t offset) {
	return addsat(chdata, offset);
}

int8_t mso_Invert(int8_t chdata) {
	return subsat(0, chdata);
}

void mso_MathApply(int8_t *ch1data, uint8_t ch1cfg, int8_t *ch2data, uint8_t ch2cfg) {
    int8_t temp1, temp2;

    temp1 = *ch1data;
    temp2 = *ch2data;
    if (testbit(ch1cfg, MAIN_CHANNEL_CONTROL_Math)) {
	    if (testbit(ch1cfg, MAIN_CHANNEL_CONTROL_Subtract)) {
			*ch1data = subsat(temp1, temp2);		// CH1 - CH2
	    } else if (testbit(ch1cfg, MAIN_CHANNEL_CONTROL_Multiply)) {
			*ch1data = fmuls8(temp1, temp2);		// CH1 * CH2
		}
    }
    if (testbit(ch2cfg, MAIN_CHANNEL_CONTROL_Math)) {
	    if (testbit(ch2cfg, MAIN_CHANNEL_CONTROL_Subtract)) {
			*ch2data = subsat(temp2, temp1);		// CH2 - CH1
		} else if (testbit(ch2cfg, MAIN_CHANNEL_CONTROL_Multiply)){
			*ch2data = fmuls8(temp1, temp2);		// CH1 * CH2
		}
    }
}

int8_t mso_ElasticApply(int8_t prevdata, int8_t newdata) {
    return average15(prevdata, newdata);
}

void mso_DMALogicChannelDelay(uint8_t samplingrate, uint8_t delay) {
	uint8_t i;	

	if (samplingrate < SAMPLING_RATE_32ksps) {     
		for (i = 0; i < delay; i++) {
			DMA.CH2.CTRLA |= DMA_CH_TRFREQ_bm;
			asm volatile("nop" : :);
		}
	}
}

uint8_t mso_CaptureBufferPreTrigger(uint8_t samplingrate, uint8_t chbuffersize) {
	uint16_t t, i;
	uint8_t result = 0;

	t = DATA_BUFFER_BLOCK_SIZE * chbuffersize;
	while((!testbit(DMA.CH0.CTRLB, DMA_CH_TRNIF_bp)) && (!testbit(DMA.CH1.CTRLB, DMA_CH_TRNIF_bp)) && (!testbit(DMA.CH2.CTRLB, DMA_CH_TRNIF_bp)) && (t != 0)) {
		for (i = 0; i < (2 << samplingrate); i++) {
			_delay_us(1);
		}
		t--;
		if (testbit(GPIO_Action, MSO_TRIGGER_STATUS_Break)) break;
	}
	if (t == 0) {
		setbit(result, MSO_ERROR_CaptureTimeout);
	} 
	return result;
}

void mso_WaitForSampleComplet(uint8_t samplingrate) {
	uint8_t i;
	
	// ATXmega-AU_Manual.pdf, ADC Clock and Conversion Timing p.346
	if (samplingrate > SAMPLING_RATE_32ksps) {
		i = 2;
	} else {
		i = (1 << samplingrate) + 2;
	}
	while (i--) {
		_delay_us(16);
	}
}

void mso_DMASamplingStart(uint8_t samplingrate) {
	if (samplingrate < SAMPLING_RATE_32ksps) {
		ADCA.CTRLB |= ADC_FREERUN_bm;
		ADCB.CTRLB |= ADC_FREERUN_bm;
		EVSYS.CH0MUX = mso_CHDpreskaler(samplingrate);
	} else TCE0.CTRLA = TC_CLKSEL_DIV8_gc;
}

void mso_DMASamplingStop(void) {
	ADCA.CTRLB &= ~(ADC_FREERUN_bm);
	ADCB.CTRLB &= ~(ADC_FREERUN_bm);
	TCE0.CTRLA = TC_CLKSEL_OFF_gc;
	_delay_us(80);
}

uint16_t mso_DMARestToTransfer(uint8_t chctrl) {
	if (testbit(chctrl, MAIN_CHANNEL_CH1)) return DMA.CH0.TRFCNT;
	if (testbit(chctrl, MAIN_CHANNEL_CH2)) return DMA.CH1.TRFCNT;
	if (testbit(chctrl, MAIN_CHANNEL_CHD)) return DMA.CH2.TRFCNT;
	return 0;
}

void mso_TriggerHold(uint16_t delay) {
	while (delay != 0) {
		if (testbit(GPIO_Request, MAIN_REQUEST_Update)) break;
		_delay_ms(1); 
		delay--;
	}
}

void mso_ConfigApply(config_t *mainconfig, uint8_t samplingrate) {
	mso_TimersInit(mainconfig->autotrigger);
	adc_InitResolution8bit();
	mso_SamplingRateApply(samplingrate, &mso_frameidx, &mso_slowsamplecount);
 }

void mso_TimersInit(uint8_t timeout) {
	TCE0.CTRLA = TC_CLKSEL_OFF_gc;				
	TCE0.CTRLB = 0;
	TCE0.INTCTRLA = 0;
	TCC1.CTRLA = TC_CLKSEL_OFF_gc;
	TCC1.CTRLB = 0;
	TCC0.CTRLA = TC_CLKSEL_OFF_gc;
	TCC0.CTRLFSET = TC_CMD_RESET_gc;
	TCC0.CTRLE = TC_BYTEM_NORMAL_gc;
	TCC0.PER = timeout;
}

void mso_TimersStop(void) {
	TCC1.CTRLA = TC_CLKSEL_OFF_gc;
	EVSYS.CH0MUX = 0;
}

void mso_DisableAutoTrigger(void) {
	TCC0.INTCTRLA &= ~TC2_LUNFINTLVL_LO_gc;
	TCC0.CTRLA = TC_CLKSEL_OFF_gc ;
}

void mso_EnableAutoTrigger(void) {
	TCC0.CTRLA = TC_CLKSEL_EVCH7_gc;
	TCC0.INTFLAGS |= TC0_OVFIF_bm;
	TCC0.INTCTRLA |= TC0_OVFINTLVL0_bm;
}

uint8_t mso_CHDpreskaler(uint8_t samplingrate) {
	return (0x80 | (samplingrate + 4));
}

void mso_SamplingRateApply(uint8_t samplingrate, volatile uint16_t *index, uint16_t *samplecount) {
	if (samplingrate >= SAMPLING_RATE_800sps) {
		TCE0.CTRLA = TC_CLKSEL_DIV4_gc;
		if (samplingrate == SAMPLING_RATE_800sps) TCE0.PER = 4999;
		else TCE0.PER = 6249;
		*samplecount = (uint16_t)pgm_read_word_near(mso_counts + samplingrate - 11);
		#if PORTE_PIN7_TEST_CLOCK_ENABLE == 0
			EVSYS.CH0MUX = EVSYS_CHMUX_TCE0_OVF_gc;
		#endif
	    ADCA.PRESCALER = ADC_PRESCALER_DIV64_gc;
	    ADCB.PRESCALER = ADC_PRESCALER_DIV64_gc;
	    ADCA.EVCTRL = ADC_EVACT_CH0_gc;
	    ADCB.EVCTRL = ADC_EVACT_CH0_gc;
	} else if (samplingrate >= SAMPLING_RATE_32ksps) {					
		*index = 0;
		TCE0.PERBUF = (uint16_t)pgm_read_word_near(mso_TCE0value + samplingrate - 6);
		#if PORTE_PIN7_TEST_CLOCK_ENABLE == 0
			EVSYS.CH0MUX = EVSYS_CHMUX_TCE0_OVF_gc;
		#endif
	    ADCA.PRESCALER = ADC_PRESCALER_DIV64_gc;
	    ADCB.PRESCALER = ADC_PRESCALER_DIV64_gc;
	    ADCA.EVCTRL = ADC_EVACT_CH0_gc;
		ADCB.EVCTRL = ADC_EVACT_CH0_gc;
	} else {
		*index = 0;
		TCE0.CTRLA = 0;
		ADCA.PRESCALER = samplingrate + 2;
		ADCA.EVCTRL = 0x00;
		ADCB.PRESCALER = samplingrate + 2;
		ADCB.EVCTRL = 0x00;
	}
}

void mso_SlowSamplingControl(volatile uint16_t *index) {
	(*index)++;
    if (*index >= running.chbufferblocks * DATA_BUFFER_BLOCK_SIZE) *index = 0;
    if (*index == 0) setbit(GPIO_Status, MAIN_STATUS_Acquired);
    if (!testbit(GPIO_Mixed, MAIN_MSO_MODE_Roll)) {
		if (*index == 0) {
			mso_SlowSamplingStop();
		    if (testbit(GPIO_Action, MSO_TRIGGER_STATUS_Stop)) return;
		}
	}
    else if (testbit(GPIO_Action, MSO_TRIGGER_STATUS_Stop)) mso_SlowSamplingStop();
}

void mso_SlowSamplingStart(void) {
	TCE0.INTCTRLA = TC_OVFINTLVL_HI_gc;
}

void mso_SlowSamplingStop(void) {
	TCE0.INTCTRLA = 0;
}

void mso_Stop(void) {
	mso_DMASamplingStop();
	mso_TimersStop();
	adc_DMAStop();
	mso_SlowSamplingStop();
	clrbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered);
}

// Slow sampling control 1600Hz or 1280Hz. Send data directly to USB buffer
ISR(TCE0_OVF_vect) {
    static uint16_t count = 0;
    static int32_t sumch1, sumch2;
    int8_t ch1, ch2;
	
	if (testbit(GPIO_CH1ctrl, MAIN_CHANNEL_CONTROL_Enable)) {
		ch1 = adc_ReadChannel8bit(&ADCA);
		ch1 = mso_OffsetApply(ch1, usbbuffers.mmch1->offset);
		if (testbit(GPIO_CH1ctrl, MAIN_CHANNEL_CONTROL_Invert)) {
			ch1 = mso_Invert(ch1);
		}
	}
	if (testbit(GPIO_CH2ctrl, MAIN_CHANNEL_CONTROL_Enable)) { 
		ch2 = adc_ReadChannel8bit(&ADCB); 
		ch2 = mso_OffsetApply(ch2, usbbuffers.mmch2->offset);
		if (testbit(GPIO_CH2ctrl, MAIN_CHANNEL_CONTROL_Invert)) {
			ch2 = mso_Invert(ch2);
		}
	}
	if (testbit(GPIO_CH1ctrl, MAIN_CHANNEL_CONTROL_Enable) && testbit(GPIO_CH2ctrl, MAIN_CHANNEL_CONTROL_Enable)) {
		mso_MathApply(&ch1, GPIO_CH1ctrl, &ch2, GPIO_CH2ctrl);
	}
    sumch1 += ch1;
    sumch2 += ch2;
    count++;
    if (count >= mso_slowsamplecount) {
        count = 0;
		usbbuffers.config->index = mso_frameidx;
		setbit(GPIO_Request, MAIN_REQUEST_USB_Send);
		if (testbit(GPIO_CH1ctrl, MAIN_CHANNEL_CONTROL_Enable)) {
			if (testbit(GPIO_CH1ctrl, MAIN_CHANNEL_CONTROL_Average)) ch1 = sumch1 / mso_slowsamplecount;
			if (testbit(GPIO_Mixed, MAIN_MSO_MODE_Elastic)) {
				ch1 = mso_ElasticApply(*(usbbuffers.ch1 + mso_frameidx), ch1);
			}
			*(usbbuffers.ch1 + mso_frameidx) = ch1;
		}
		if (testbit(GPIO_CH2ctrl, MAIN_CHANNEL_CONTROL_Enable)) {
			if (testbit(GPIO_CH2ctrl, MAIN_CHANNEL_CONTROL_Average)) ch2 = sumch2 / mso_slowsamplecount;
			if (testbit(GPIO_Mixed, MAIN_MSO_MODE_Elastic)) {
				ch2 = mso_ElasticApply(*(usbbuffers.ch2 + mso_frameidx), ch2);
			}
			*(usbbuffers.ch2 + mso_frameidx) = ch2;
		}
        if (testbit(GPIO_CHDctrl, MAIN_CHANNEL_CONTROL_Enable)) { 
			*(usbbuffers.chd + mso_frameidx) = VPORT2.IN;
		}
        sumch1 = 0; 
		sumch2 = 0;
		mso_SlowSamplingControl(&mso_frameidx);
    }
}

// Interrupt for auto trigger
ISR(TCC0_OVF_vect) {
	mso_DisableAutoTrigger();
    setbit(GPIO_Request, MAIN_REQUEST_Update);
    setbit(GPIO_Action, MSO_TRIGGER_STATUS_Triggered);	
}

