/*
 * multimeter.c part of XMEGA Oscilloscope
 *
 * Created: 2016-12-16 09:47:32
 * Atmel Studio 6 (Version: 6.2.1563 - Service Pack 2)
 *  Author: Krzysztof Markiewicz
 *  www.obbo.pl
 *
 * This program is distributed under the terms of the GNU General Public License
 */ 

#include <math.h>
#include "multimeter.h"
#include "calibration.h"
#include "input.h"
#include "adc.h"
#include "bitlib.h"



typedef struct {
	uint16_t first;
	uint16_t second;
	uint16_t last;
} mmzeros_t;


uint32_t mm_MeasureVACFrequency(int8_t* buffer, const uint16_t len, int16_t min, int16_t max, mmzeros_t* cross);
void mm_MeasureFrequencyInit(void);
void mm_MeasurePeriodInit(void);
void mm_MeasureFrequenceStop(void);
void mm_MeasurePeriodStop(const uint8_t trigersource);
void mm_CountPulsesStop(void);
void mm_CountPulsesGatedStop(const uint8_t trigersource, const uint8_t gate);
uint8_t mm_MeasureTrigerSource(const uint8_t trigersource);
void mm_MeasureTrigerConfig(const uint8_t trigersource, uint8_t edge, bool invert);
void mm_TC0ClearIntFlags(TC0_t *timer);
void mm_TC1ClearIntFlags(TC1_t *timer);
void mm_RestoreLogicPortConfiguration(void);


// The result is the sum of 256 measurements with a resolution of 12 bit
void mm_MeasureSeriesVoltage12bit(const uint8_t ch1cfg, const uint8_t ch1gain, int32_t *ch1vdc, const uint8_t ch2cfg, const uint8_t ch2gain, int32_t *ch2vdc, bool calibrate) {
	uint8_t i = 0;
	
	adc_DMAStop();
	adc_InitResolution12bit();
	adc_InitGain(ch1gain, ch2gain);
	_delay_ms(1);
 	data_StoreCHDataGain(ch1gain, ch2gain);
	adc_InitTrigered();
	*ch1vdc = 0;
	*ch2vdc = 0;
	do {					
		ADCA.CH0.CTRL |= ADC_CH_START_bm;
		ADCB.CH0.CTRL |= ADC_CH_START_bm;
		// ATXmega-AU_Manual.pdf, ADC Clock and Conversion Timing p.346
		_delay_us(170);
		*ch1vdc = *ch1vdc - (int16_t)(ADCA.CH0.RES);
		*ch2vdc = *ch2vdc - (int16_t)(ADCB.CH0.RES);
	} while(++i);
	if (!calibrate) {
		*ch1vdc -= (int32_t)eeprom_read_dword((uint32_t *)&adc_offset12CH1[calibration_ADCOffsetCell(ch1cfg, ch1gain)]);
		*ch2vdc -= (int32_t)eeprom_read_dword((uint32_t *)&adc_offset12CH2[calibration_ADCOffsetCell(ch2cfg, ch2gain)]);
	}
}

// The result is the sum of 8 measurements with a resolution of 12 bit
void mm_MeasureVDC(const uint8_t ch1cfg, const uint8_t ch1gain, channela_t *ch1vdc, const uint8_t ch2cfg, const uint8_t ch2gain, channela_t *ch2vdc) {
	int32_t ch1vdc32 = 0, ch2vdc32 = 0;

	mm_MeasureSeriesVoltage12bit(ch1cfg, ch1gain, &ch1vdc32, ch2cfg, ch2gain, &ch2vdc32, false);
	ch1vdc->aver = (int16_t)(ch1vdc32 / 32);
	ch2vdc->aver = (int16_t)(ch2vdc32 / 32);
}
 
// The result is the sum of 256 measurements with a resolution of 8 bit
void mm_MeasureSeriesVoltage8bit(const uint8_t ch1cfg, const uint8_t ch1gain, int16_t *ch1vdc, const uint8_t ch2cfg, const uint8_t ch2gain, int16_t *ch2vdc, bool calibrate) {
	int8_t ch1offset, ch2offset;
	uint8_t i = 0;
	 
	adc_DMAStop();
	adc_InitResolution8bit();
	adc_InitGain(ch1gain, ch2gain);
	_delay_ms(1);
 	data_StoreCHDataGain(ch1gain, ch2gain);
	adc_InitTrigered();
	*ch1vdc = 0;
	*ch2vdc = 0;
	ch1offset = (int8_t)eeprom_read_byte((uint8_t *)&adc_offset8CH1[calibration_ADCOffsetCell(ch1cfg, ch1gain)]);
	ch2offset = (int8_t)eeprom_read_byte((uint8_t *)&adc_offset8CH2[calibration_ADCOffsetCell(ch2cfg, ch2gain)]);
	do {
		ADCA.CH0.CTRL |= ADC_CH_START_bm;
		ADCB.CH0.CTRL |= ADC_CH_START_bm;
		// ATXmega-AU_Manual.pdf, ADC Clock and Conversion Timing p.346
		_delay_us(150);
		*ch1vdc = *ch1vdc - (int8_t)(ADCA.CH0.RES);
		*ch2vdc = *ch2vdc - (int8_t)(ADCB.CH0.RES);
		if (!calibrate) {
			*ch1vdc -= ch1offset;
			*ch2vdc -= ch2offset;
		}
	} while(++i);
} 
 
 void mm_MeasureClearResult(channela_t *channel) {
	channel->max = 0;
	channel->min = 0;
	channel->rms = 0;
	channel->aver = 0;
	channel->frequency = 0;
	channel->cross1 = 64;
	channel->cross2 = 192;
 }

// Work on USB buffer channel data, return frequency as count of samples for 100 periods
uint32_t mm_MeasureVACFrequency(int8_t* buffer, const uint16_t length, int16_t min, int16_t max, mmzeros_t* cross) {
	#define BORDER	5
	uint8_t step = 0, crosses = 0; 
	int8_t half;
	uint16_t i, idx[3], periods = 0;
	uint32_t frequency = 0;
	bool dirup = false;
	
	cross->first = 0;
	cross->second = 0;
	cross->last = length;
	if ((max - min) < 32) return 0;
	half = (int8_t)((min + max) / 2);
	min = (min * 10 + max * 6) / 16;
	max = (min * 6 + max * 10) / 16;
	for (i = BORDER; i < (length - BORDER); i++) {
		switch (step) {
			case 0:
				if (buffer[i] < min) {
					step = 1; 
					dirup = true;
				} else if (buffer[i] > max) {
					step = 1;
					dirup = false;
				}
				break;
			case 1:
				if ((dirup && (buffer[i] > min)) || (!dirup && (buffer[i] < max))) {
					step = 2;
				}
			case 2:
				if ((dirup && (buffer[i] > half)) || (!dirup && (buffer[i] < half))) {
					step = 3;
					idx[crosses] = i;
					if (crosses < 2) crosses++;
					periods++;
				}
				break;
			case 3:
				if ((dirup && (buffer[i] < min)) || (!dirup && (buffer[i] > max))) {
					step = 1;
				}
				break;
		}
		if (periods > (length >> 2)) return 0; 
	}
	periods--;
	if (periods < 1) return 0;
	else if (periods == 1) idx[2] = idx[1];
	cross->first = idx[0];
	cross->second = idx[1];
	cross->last = idx[2];
	frequency = ((100 * (int32_t)(idx[2] - idx[0])) / periods);
	return frequency;
}

// Work on USB buffer data
void mm_MeasureVAC(int8_t *buffer, const uint16_t length, channela_t *channel) {
	int8_t tmpmin, tmpmax;
	mmzeros_t idx;
	uint16_t i;
	int32_t tmpaverage = 0, tmprms = 0;
	
	tmpmin = buffer[0];
	tmpmax = buffer[0];
	for (i = 0; i < length; i++) {
		if (buffer[i] > tmpmax) tmpmax = buffer[i];
		if (buffer[i] < tmpmin) tmpmin = buffer[i];
	}
	channel->min = tmpmin;
	channel->max = tmpmax;
	channel->frequency = mm_MeasureVACFrequency(buffer, length, tmpmin, tmpmax, &idx);
	channel->cross1 = idx.first;
	channel->cross2 = idx.second;
	for (i = idx.first; i < idx.last; i++) {
		tmpaverage += buffer[i];
		tmprms += buffer[i] * buffer[i];
	}
	// 128 measurements with a resolution of 8 bit
	channel->rms = (int16_t)(128 * sqrt(tmprms / (idx.last - idx.first)));
	// 128 measurements with a resolution of 8 bit
	channel->aver = (int16_t)((tmpaverage * 128) / (idx.last - idx.first)); 
}

void mm_TC0ClearIntFlags(TC0_t *timer) {
	timer->INTFLAGS = 0xff;
}

void mm_TC1ClearIntFlags(TC1_t *timer) {
	timer->INTFLAGS = 0xff;
}

void mm_MeasureFrequencyInit(void) {
    EVSYS.CH4MUX = EVSYS_CHMUX_TCC0_OVF_gc;
    EVSYS.CH5MUX = EVSYS_CHMUX_TCC1_OVF_gc;
	// Frequency Hi16
	TCE0.CTRLA = TC_CLKSEL_OFF_gc; 
	TCE0.CTRLFSET = TC_CMD_RESET_gc; 
	TCE0.CNT = 0;
	TCE0.CTRLA = TC_CLKSEL_EVCH5_gc;
	TCE0.CTRLD = TC_EVACT_CAPT_gc | TC1_EVDLY_bm | TC_EVSEL_CH4_gc;
	TCE0.CTRLB = TC0_CCAEN_bm; 
	// Frequency Lo16
	TCC1.CTRLA = TC_CLKSEL_OFF_gc;
	TCC1.CTRLFSET = TC_CMD_RESET_gc;
	TCC1.CNT = 0;
	TCC1.CTRLD = TC_EVACT_CAPT_gc | TC_EVSEL_CH4_gc;
	TCC1.CTRLB = TC1_CCAEN_bm;
	// Gate
	TCC0.CTRLA = TC_CLKSEL_OFF_gc;
	TCC0.CTRLFSET = TC_CMD_RESET_gc;
	TCC0.CTRLE = TC_BYTEM_NORMAL_gc;
	TCC0.CNT = 0;
}

// Base on Atmel AVR1617: Frequency Measurement with Atmel AVR XMEGA Family Devices
void mm_MeasureFrequency(config_t *measuremode, const uint8_t trigersource, channela_t *result, bool fastmode) {
	uint8_t clock;
	uint32_t f = 0;
	
	if (fastmode) {
		clock = TC_CLKSEL_DIV256_gc;
	} else {
		clock = TC_CLKSEL_DIV1024_gc;
	}
	EVSYS.CH2MUX = mm_MeasureTrigerSource(trigersource);
	mm_MeasureFrequencyInit();
	TCC0.PER = 31249;
	mm_TC0ClearIntFlags(&TCC0);
	asm volatile("nop" : :);
	TCC0.CTRLA = clock;
	TCC1.CTRLA = TC_CLKSEL_EVCH2_gc;
	while (!testbit(TCC0.INTFLAGS, TC1_OVFIF_bp)) {
		if (measuremode->mainmode != MAIN_MODE_MeterFrequency) {
			mm_MeasureFrequenceStop();
			return;
		}
	}
	asm volatile("nop" "\n\t" "nop" : :);
	f = TCE0.CCA;
	f = (f << 16) + TCC1.CCA;
	if (fastmode) f = f << 2;
	result->frequency = f;
	mm_MeasureFrequenceStop();
}

void mm_MeasureFrequenceStop(void) {
	EVSYS.CH2MUX = 0;
	EVSYS.CH4MUX = 0;
	EVSYS.CH5MUX = 0;
	TCC0.CTRLA = TC_CLKSEL_OFF_gc;
	mm_TC0ClearIntFlags(&TCC0);
}

void mm_MeasurePeriodInit(void) {
	EVSYS.CH5MUX = EVSYS_CHMUX_TCC1_OVF_gc;
	EVSYS.CH4MUX = EVSYS_CHMUX_TCE1_OVF_gc;
	// Period Hi16
	TCE0.CTRLA = TC_CLKSEL_OFF_gc;
	TCE0.CTRLFSET = TC_CMD_RESET_gc;
	TCE0.CNT = 0;
	TCE0.CTRLA = TC_CLKSEL_EVCH5_gc;
	TCE0.CTRLD = TC_EVACT_FRQ_gc | TC1_EVDLY_bm | TC_EVSEL_CH2_gc;
	TCE0.CTRLB = TC0_CCAEN_bm;
	// Period Lo16
	TCC1.CTRLA = TC_CLKSEL_OFF_gc;
	TCC1.CTRLFSET = TC_CMD_RESET_gc;
	TCC1.CNT = 0;
	TCC1.CTRLA = TC_CLKSEL_DIV1_gc;
	TCC1.CTRLD = TC_EVACT_FRQ_gc | TC_EVSEL_CH2_gc;
	TCC1.CTRLB = TC1_CCAEN_bm;
	// Duty Hi16
	TCF0.CTRLA = TC_CLKSEL_OFF_gc;
	TCF0.CTRLFSET = TC_CMD_RESET_gc;
	TCF0.CNT = 0;
	TCF0.CTRLD = TC_EVACT_PW_gc | TC1_EVDLY_bm | TC_EVSEL_CH2_gc;
	TCF0.CTRLB = TC0_CCAEN_bm;
	TCF0.CTRLA = TC_CLKSEL_EVCH4_gc;
	// Duty Lo16
	TCE1.CTRLA = TC_CLKSEL_OFF_gc;
	TCE1.CTRLFSET = TC_CMD_RESET_gc;
	TCE1.CNT = 0;
	TCE1.CTRLD = TC_EVACT_PW_gc | TC_EVSEL_CH2_gc;
	TCE1.CTRLB = TC1_CCAEN_bm;
	TCE1.CTRLA = TC_CLKSEL_DIV1_gc;
}

uint8_t mm_MeasurePeriod(config_t *measuremode, const uint8_t trigersource, channela_t *result, channela_t *duty) {
	uint32_t f = 0;
	uint8_t count = 0, status = 0;

	mm_MeasureTrigerConfig(trigersource, INPUT_PORT_SENSE_EDGE_Bothedges, false);
	EVSYS.CH2MUX = mm_MeasureTrigerSource(trigersource);
    mm_MeasurePeriodInit();
	mm_TC0ClearIntFlags(&TCE0);
	asm volatile("nop" : :);
	while (count < 2) {
		mm_TC1ClearIntFlags(&TCC1);
		asm volatile("nop" : :);
		while (!testbit(TCC1.INTFLAGS, TC1_CCAIF_bp)) {
			if ((measuremode->mainmode != MAIN_MODE_MeterPeriod) || (testbit(GPIO_Action, MSO_TRIGGER_STATUS_Break))) {
				break;
			}
			if (testbit(TCE0.INTFLAGS, TC1_OVFIF_bp)) {
				result->frequency = 0xffffffff;
				mm_MeasurePeriodStop(trigersource);
				return MM_ERROR_MEASURE_OutOfRange;
			}
		}
		count++;
	}
	asm volatile("nop" "\n\t" "nop" : :);
	f = TCE0.CCA;
	f = (f << 16) + TCC1.CCA;
	result->frequency = f;
	f = TCF0.CCA;
	f = (f << 16) + TCE1.CCA;
	duty->frequency = f;
	mm_MeasurePeriodStop(trigersource);
	return status;
}

void mm_MeasurePeriodStop(const uint8_t trigersource) {
	EVSYS.CH2MUX = 0;
	EVSYS.CH4MUX = 0;
	EVSYS.CH5MUX = 0;
	TCC1.CTRLA = TC_CLKSEL_OFF_gc;
	TCE1.CTRLA = TC_CLKSEL_OFF_gc;
	mm_TC0ClearIntFlags(&TCC0);
	mm_RestoreLogicPortConfiguration(); 
}

uint8_t mm_CountPulses(config_t *measuremode, const uint8_t trigersource, channela_t *result) { 
	uint32_t f = 0;
	uint8_t status = 0;
	
	mm_MeasureTrigerConfig(trigersource, INPUT_PORT_SENSE_EDGE_Rising, false);
	EVSYS.CH2MUX = mm_MeasureTrigerSource(trigersource);
	mm_MeasureFrequencyInit();
	TCC1.CTRLA = TC_CLKSEL_EVCH2_gc;
	while (!testbit(GPIO_MeterFrequency, METER_FREQUENCY_Start)) {
		if (measuremode->mainmode != MAIN_MODE_MeterCount) {
			mm_CountPulsesStop();
			return status;
		}
	}
	clrbit(GPIO_MeterFrequency, METER_FREQUENCY_Stop);
	TCE0.CTRLFSET = TC_CMD_RESTART_gc;
	TCC1.CTRLFSET = TC_CMD_RESTART_gc;
	while (!testbit(GPIO_MeterFrequency, METER_FREQUENCY_Stop)) {
		if (measuremode->mainmode != MAIN_MODE_MeterCount) {
			mm_CountPulsesStop();
			return status;
		}
		if (testbit(TCE0.INTFLAGS, TC1_OVFIF_bp)) {
			result->frequency = 0xffffffff;
			mm_CountPulsesStop();
			return MM_ERROR_MEASURE_OutOfRange;
		}
		f = TCE0.CNT;
		f = (f << 16) + TCC1.CNT;
		result->frequency = f;
	}
	EVSYS.STROBE = 0b00010000;
	asm volatile("nop" "\n\t" "nop" : :);
	f = TCE0.CCA;
	f = (f << 16) + TCC1.CCA;
	result->frequency = f;
	mm_CountPulsesStop();
	return status;
}

void mm_CountPulsesStop(void) {
	EVSYS.CH2MUX = 0;
	EVSYS.CH4MUX = 0;
	EVSYS.CH5MUX = 0;
	mm_RestoreLogicPortConfiguration();
	clrbit(GPIO_MeterFrequency, METER_FREQUENCY_Start);
	clrbit(GPIO_MeterFrequency, METER_FREQUENCY_Stop);
}

uint8_t mm_CountPulsesWithGate(config_t *measuremode, const uint8_t trigersource, const uint8_t gate, channela_t *result, uint8_t gatemode) {
	uint32_t f = 0;
	uint8_t count = 0, status = 0;
	
	mm_MeasureFrequencyInit();
	TCC1.CTRLA = TC_CLKSEL_EVCH2_gc;
	if (testbit(gatemode, METER_FREQUENCY_GATE_Period)) {
		TCE0.CTRLD = TC_EVACT_FRQ_gc | TC1_EVDLY_bm | TC_EVSEL_CH4_gc;
		TCC1.CTRLD = TC_EVACT_FRQ_gc | TC_EVSEL_CH4_gc;
	} else {
		TCE0.CTRLD = TC_EVACT_PW_gc | TC1_EVDLY_bm | TC_EVSEL_CH4_gc;
		TCC1.CTRLD = TC_EVACT_PW_gc | TC_EVSEL_CH4_gc;
	}
	mm_MeasureTrigerConfig(gate, INPUT_PORT_SENSE_EDGE_Bothedges, false);
	EVSYS.CH4MUX = mm_MeasureTrigerSource(gate);
	mm_MeasureTrigerConfig(trigersource, INPUT_PORT_SENSE_EDGE_Rising, false);
	EVSYS.CH2MUX = mm_MeasureTrigerSource(trigersource);
	mm_TC0ClearIntFlags(&TCE0);
	asm volatile("nop" : :);
	while (count < 2) {
		mm_TC1ClearIntFlags(&TCC1);
		asm volatile("nop" : :);
		while (!testbit(TCC1.INTFLAGS, TC1_CCAIF_bp)) {
			if ((measuremode->mainmode != MAIN_MODE_MeterGate) || (testbit(GPIO_Action, MSO_TRIGGER_STATUS_Break))) {
				break;
			}
			if (testbit(TCE0.INTFLAGS, TC1_OVFIF_bp)) {
				result->frequency = 0xffffffff;
				mm_CountPulsesGatedStop(trigersource, gate);
				return MM_ERROR_MEASURE_OutOfRange;
			}
		}
		count++;
	}
	asm volatile("nop" "\n\t" "nop" : :);
	mm_CountPulsesGatedStop(trigersource, gate);
	f = TCE0.CCA;
	f = (f << 16) + TCC1.CCA;
	result->frequency = f;
	return status;
}

void mm_CountPulsesGatedStop(const uint8_t trigersource, const uint8_t gate) {
	EVSYS.CH4MUX = 0;
	EVSYS.CH2MUX = 0;
	mm_RestoreLogicPortConfiguration();
}

uint8_t mm_MeasureTrigerSource(const uint8_t trigersource) {
	if ((trigersource >= MAIN_TRIGGER_SOURCE_CHD_Bit0) && (trigersource <= MAIN_TRIGGER_SOURCE_CHD_Bit7)) {
		return (EVSYS_CHMUX_PORTC_PIN0_gc + (trigersource - ADC_CHANNEL_ANALOG_COUNT));
	} else if (trigersource == MAIN_TRIGGER_SOURCE_EXT) {
		return (EVSYS_CHMUX_PORTD_PIN1_gc);
	} else if (trigersource == MAIN_TRIGGER_SOURCE_SCHMITT) {
		return (EVSYS_CHMUX_PORTD_PIN0_gc);
	} else {
		return (EVSYS_CHMUX_PORTD_PIN1_gc);
	}
}

void mm_MeasureTrigerConfig(const uint8_t trigersource, uint8_t edge, bool invert) {
	if ((trigersource >= MAIN_TRIGGER_SOURCE_CHD_Bit0) && (trigersource <= MAIN_TRIGGER_SOURCE_CHD_Bit7)) {
		input_LogicPinConfig(&PORTC, (trigersource - ADC_CHANNEL_ANALOG_COUNT), invert, edge);
	} else if (trigersource == MAIN_TRIGGER_SOURCE_EXT) {
		input_LogicPinConfig(&PORTD, INPUT_EXT_PIN, invert, edge);
	} else if (trigersource == MAIN_TRIGGER_SOURCE_SCHMITT) {
		input_LogicPinConfig(&PORTD, INPUT_SCHMITT_PIN, invert, edge);
	} else {
		input_LogicPinConfig(&PORTD, INPUT_EXT_PIN, invert, edge);
	}
}

void mm_RestoreLogicPortConfiguration(void) {
	input_LogicPortConfig(&PORTC, testbit(GPIO_CHDctrl, MAIN_CHANNEL_CONTROL_Invert), INPUT_PORT_SENSE_EDGE_Rising);
	input_LogicPinConfig(&PORTD, INPUT_SCHMITT_PIN, false, INPUT_PORT_SENSE_EDGE_Rising);
	input_LogicPinConfig(&PORTD, INPUT_EXT_PIN, false, INPUT_PORT_SENSE_EDGE_Rising);
}