/*
 * sniffer.c part of XMEGA Oscilloscope
 *
 * Created: 2016-12-16
 * Atmel Studio 6 (Version: 6.2.1563 - Service Pack 2)
 *  Author: Krzysztof Markiewicz
 *  www.obbo.pl
 *
 * This program is distributed under the terms of the GNU General Public License
 */ 

#include <stdint.h>
#include <avr/interrupt.h>
#include <stdbool.h>
#include <util/delay.h>
#include "sniffer.h"
#include "input.h"
#include "usb.h"
#include "bitlib.h"

#define I2C_SDA_PIN		0
#define I2C_SCL_PIN		1
#define SPI_SS_PIN		4
#define SPI_MOSI_PIN	5
#define SPI_MISO_PIN	6	
#define SPI_SCK_PIN		7
#define SCL				(unsigned char)(VPORT2.IN & _BV(I2C_SCL_PIN))
#define SDA				(unsigned char)(VPORT2.IN & _BV(I2C_SDA_PIN))
#define EVSYS_CH2_bm	0b00000100 


void sniffer_DMAInit(uint8_t protocol, DMA_CH_t *dma);
void sniffer_DMAEnable(DMA_CH_t *dma, uint8_t sniff);
uint16_t sniffer_DMALength(uint8_t protocol);
void sniffer_DMAStart(uint8_t protocol, sniffbuffer_t buffer, uint8_t option, uint8_t sniff);
void sniffer_DMAStopChannel(DMA_CH_t *dma);
void sniffer_DMAStop(void);
void sniffer_ClearBuffer(uint8_t protocol, sniffbuffer_t buffer, uint8_t option, uint8_t sniff);
uint8_t sniffer_USARTDataLengthToUSARTRegister(uint8_t length);
uint8_t sniffer_USARTFlushBuffer(USART_t *port);
void sniffer_USARTInit(USART_t *port, uint8_t baud, uint8_t databits, uint8_t option);
void sniffer_USARTInitProtocol(sniffconfig_t *config, uint8_t bitrate, uint8_t option);
void sniffer_I2CSetIntLevel(void);
void sniffer_I2CInit(sniffbuffer_t buffer);
void sniffer_SPIInitPort(PORT_t *port, uint8_t option);
void sniffer_SPIInit(uint8_t option);
void sniffer_SPIStartPeripheral(SPI_t *spi, uint8_t option);
void sniffer_SPIStart(uint8_t option);
void sniffer_PortIntFlagsClear(void);
void sniffer_InitTimeoutTimer(void);
void sniffer_ParallelInit(uint8_t option, uint8_t edge);
void sniffer_ParallelStart(void);
void sniffer_PowerOnPeripherials(void);
void sniffer_PowerOffPeripherials(void);
void sniffer_USARTStopPort(USART_t *port);
void sniffer_USARTStop(void);
void sniffer_SPIStopPort(SPI_t *spi);
void sniffer_SPIStop(void);
void sniffer_I2CStop(void);
void sniffer_Stop(void);


volatile uint8_t rxdata, txdata, dxdata;
volatile bool i2cstart = false;
volatile bool i2cfirst;


// export definition to Xlab oscyloskopx.h
#define SNIFFER_VALID_Rx_DATA_Bit		0b00000001
#define SNIFFER_VALID_Tx_DATA_Bit		0b00000010
#define SNIFFER_I2C_ACK					0b00000001
#define SNIFFER_I2C_Address				0b00000010
#define SNIFFER_I2C_Start				0b00000100
#define SNIFFER_I2C_RepeatedStart		0b00001000
#define SNIFFER_I2C_Stop				0b00010000
#define SNIFFER_I2C_Timeout				0b00100000
#define SNIFFER_I2C_FastStop			0b01000000
#define SNIFFER_I2C_ValidData			0b10000000
// end export definition to Xlab oscyloskopx.h

const uint8_t BAUDA[8] PROGMEM = {
    3,      // 1200
    1,      // 2400
    253,    // 4800
    245,    // 9600
    229,    // 19200
    197,    // 38400
    110,    // 57600
    23,     // 115200
};

const uint8_t BAUDB[8] PROGMEM = {
    253,    // 1200
    237,    // 2400
    220,    // 4800
    204,    // 9600
    188,    // 19200
    172,    // 38400
    168,    // 57600
    164     // 115200
};

void sniffer_DMAInit(uint8_t protocol, DMA_CH_t *dma) {
	dma->CTRLA = DMA_CH_RESET_bm;
	dma->ADDRCTRL = DMA_CH_SRCRELOAD_BURST_gc | DMA_CH_DESTRELOAD_BLOCK_gc | DMA_CH_DESTDIR_INC_gc;
	dma->TRFCNT = sniffer_DMALength(protocol);
}

void sniffer_DMAEnable(DMA_CH_t *dma, uint8_t sniff) {
	dma->CTRLA = DMA_CH_ENABLE_bm | DMA_CH_SINGLE_bm | DMA_CH_BURSTLEN_1BYTE_gc;
	if (!testbit(sniff, MAIN_SNIFFER_SingleSniff)) {
		dma->CTRLA |= DMA_CH_REPEAT_bm;
	}
}

uint16_t sniffer_DMALength(uint8_t protocol) {
	uint16_t result = 0;
	
	if (protocol == SNIFFER_PROTOCOL_USART) {
		return SNIFFER_USART_BUFFER_LENGTH;
	} else if (protocol == SNIFFER_PROTOCOL_SPI) {
		return SNIFFER_SPI_BUFFER_LENGTH;
	} else if (protocol == SNIFFER_PROTOCOL_Parallel) {
		return SNIFFER_PARALLEL_BUFFER_LENGTH;
	} else if (protocol == SNIFFER_PROTOCOL_I2C) {
		return SNIFFER_I2C_BUFFER_LENGTH;
	}
	return result;
}

void sniffer_DMAStart(uint8_t protocol, sniffbuffer_t buffer, uint8_t option, uint8_t sniff) {
	// DMA CH0 for USART Rx, SPI MOSI, Parallel, I2C Data
	sniffer_DMAInit(protocol, &DMA.CH0);
    if ((protocol == SNIFFER_PROTOCOL_USART) && (testbit(option, SNIFFER_RxTx_Synchronization))) {
		DMA.CH0.DESTADDR0 = ((uint16_t)(buffer.rx)) & 0xff;
		DMA.CH0.DESTADDR1 = (((uint16_t)(buffer.rx)) >> 8) & 0xff;
        DMA.CH0.TRIGSRC = DMA_CH_TRIGSRC_EVSYS_CH2_gc;
        DMA.CH0.SRCADDR0 = ((uint16_t)(&rxdata)) & 0xff;
        DMA.CH0.SRCADDR1 = (((uint16_t)(&rxdata)) >> 8) & 0xff;
    } else if ((protocol == SNIFFER_PROTOCOL_USART) && (!testbit(option, SNIFFER_RxTx_Synchronization))) {
		DMA.CH0.DESTADDR0 = ((uint16_t)(buffer.rx)) & 0xff;
		DMA.CH0.DESTADDR1 = (((uint16_t)(buffer.rx)) >> 8) & 0xff;
        DMA.CH0.TRIGSRC = DMA_CH_TRIGSRC_USARTC0_RXC_gc;
        DMA.CH0.SRCADDR0 = ((uint16_t)(&USARTC0.DATA)) & 0xff;
        DMA.CH0.SRCADDR1 = (((uint16_t)(&USARTC0.DATA)) >> 8) & 0xff;
	} else if (protocol == SNIFFER_PROTOCOL_SPI) {
		DMA.CH0.DESTADDR0 = ((uint16_t)(buffer.mosi)) & 0xff;
		DMA.CH0.DESTADDR1 = (((uint16_t)(buffer.mosi)) >> 8) & 0xff;
        DMA.CH0.TRIGSRC = DMA_CH_TRIGSRC_SPIC_gc;
        DMA.CH0.SRCADDR0 = ((uint16_t)(&SPIC.DATA)) & 0xff;
        DMA.CH0.SRCADDR1 = (((uint16_t)(&SPIC.DATA)) >> 8) & 0xff;
    } else if (protocol == SNIFFER_PROTOCOL_Parallel) {
		DMA.CH0.DESTADDR0 = ((uint16_t)(buffer.rx)) & 0xff;
		DMA.CH0.DESTADDR1 = (((uint16_t)(buffer.rx)) >> 8) & 0xff;
		DMA.CH0.TRIGSRC = DMA_CH_TRIGSRC_EVSYS_CH2_gc;
		DMA.CH0.SRCADDR0 = ((uint16_t)(&VPORT2.IN)) & 0xff;
		DMA.CH0.SRCADDR1 = (((uint16_t)(&VPORT2.IN)) >> 8) & 0xff;
	} else if (protocol == SNIFFER_PROTOCOL_I2C) {
		DMA.CH0.DESTADDR0 = ((uint16_t)(buffer.i2c)) & 0xff;
		DMA.CH0.DESTADDR1 = (((uint16_t)(buffer.i2c)) >> 8) & 0xff;
        DMA.CH0.TRIGSRC = DMA_CH_TRIGSRC_EVSYS_CH2_gc;
        DMA.CH0.SRCADDR0 = ((uint16_t)(&rxdata)) & 0xff;
        DMA.CH0.SRCADDR1 = (((uint16_t)(&rxdata)) >> 8) & 0xff;
	}
	sniffer_DMAEnable(&DMA.CH0, sniff);
	// DMA CH1 for USART Tx, SPI MISO, I2C ack
	sniffer_DMAInit(protocol, &DMA.CH1);
    if ((protocol == SNIFFER_PROTOCOL_USART) && (testbit(option, SNIFFER_RxTx_Synchronization))) {
		DMA.CH1.DESTADDR0 = ((uint16_t)(buffer.tx)) & 0xff;
		DMA.CH1.DESTADDR1 = (((uint16_t)(buffer.tx)) >> 8) & 0xff;
	    DMA.CH1.TRIGSRC = DMA_CH_TRIGSRC_EVSYS_CH2_gc;
	    DMA.CH1.SRCADDR0 = ((uint16_t)(&txdata)) & 0xff;
	    DMA.CH1.SRCADDR1 = (((uint16_t)(&txdata)) >> 8) & 0xff;
	} else if ((protocol == SNIFFER_PROTOCOL_USART) && (!testbit(option, SNIFFER_RxTx_Synchronization))) {
		DMA.CH1.DESTADDR0 = ((uint16_t)(buffer.tx)) & 0xff;
		DMA.CH1.DESTADDR1 = (((uint16_t)(buffer.tx)) >> 8) & 0xff;
        DMA.CH1.TRIGSRC = DMA_CH_TRIGSRC_USARTC1_RXC_gc;
        DMA.CH1.SRCADDR0 = ((uint16_t)(&USARTC1.DATA)) & 0xff;
        DMA.CH1.SRCADDR1 = (((uint16_t)(&USARTC1.DATA)) >> 8) & 0xff;
    } else if (protocol == SNIFFER_PROTOCOL_SPI) {
		DMA.CH1.DESTADDR0 = ((uint16_t)(buffer.miso)) & 0xff;
		DMA.CH1.DESTADDR1 = (((uint16_t)(buffer.miso)) >> 8) & 0xff;
		DMA.CH1.TRIGSRC = DMA_CH_TRIGSRC_SPIE_gc;
		DMA.CH1.SRCADDR0 = ((uint16_t)(&SPIE.DATA)) & 0xff;
		DMA.CH1.SRCADDR1 = (((uint16_t)(&SPIE.DATA)) >> 8) & 0xff;        
	} else if (protocol == SNIFFER_PROTOCOL_I2C) {
		DMA.CH1.DESTADDR0 = ((uint16_t)(buffer.i2cack)) & 0xff;
		DMA.CH1.DESTADDR1 = (((uint16_t)(buffer.i2cack)) >> 8) & 0xff;
		DMA.CH1.TRIGSRC = DMA_CH_TRIGSRC_EVSYS_CH2_gc;
		DMA.CH1.SRCADDR0 = ((uint16_t)(&txdata)) & 0xff;
		DMA.CH1.SRCADDR1 = (((uint16_t)(&txdata)) >> 8) & 0xff;
	}
	if ((protocol == SNIFFER_PROTOCOL_USART) || (protocol == SNIFFER_PROTOCOL_SPI) || (protocol == SNIFFER_PROTOCOL_I2C)) {
		sniffer_DMAEnable(&DMA.CH1, sniff);
	}
	// DMA CH2 for USART Dx
    if ((protocol == SNIFFER_PROTOCOL_USART) && (testbit(option, SNIFFER_RxTx_Synchronization))) {
		sniffer_DMAInit(protocol, &DMA.CH2);
		DMA.CH2.DESTADDR0 = ((uint16_t)(buffer.dx)) & 0xff;
		DMA.CH2.DESTADDR1 = (((uint16_t)(buffer.dx)) >> 8) & 0xff;
	    DMA.CH2.TRIGSRC   = DMA_CH_TRIGSRC_EVSYS_CH2_gc;
	    DMA.CH2.SRCADDR0  = ((uint16_t)(&dxdata)) & 0xff;
	    DMA.CH2.SRCADDR1  = (((uint16_t)(&dxdata)) >> 8) & 0xff;
		sniffer_DMAEnable(&DMA.CH2, sniff);
	}
}

void sniffer_DMAStopChannel(DMA_CH_t *dma) {
	uint8_t t = 100;
	
	dma->CTRLA &= ~(DMA_CH_ENABLE_bm);
	while ((dma->CTRLA & DMA_CH_ENABLE_bm) && (t != 0)) {
		_delay_us(1);
		t--;
	}
	dma->TRIGSRC = DMA_CH_TRIGSRC_OFF_gc;
	dma->TRFCNT = 0;
}

void sniffer_DMAStop(void) {
	sniffer_DMAStopChannel(&DMA.CH0);
	sniffer_DMAStopChannel(&DMA.CH1);
	sniffer_DMAStopChannel(&DMA.CH2);
}

void sniffer_ClearBuffer(uint8_t protocol, sniffbuffer_t buffer, uint8_t option, uint8_t sniff) {
	PORTC.INTCTRL = 0x00;
	sniffer_DMAStop();
	data_BufferClear8(buffer.all, data_SnifferDataLength(), 0);
	if (protocol == SNIFFER_PROTOCOL_USART) {
		sniffer_USARTFlushBuffer(&USARTC0);
		sniffer_USARTFlushBuffer(&USARTC1);
	} 
	sniffer_DMAStart(protocol, buffer, option, sniff);
	if (protocol == SNIFFER_PROTOCOL_I2C) {
		sniffer_PortIntFlagsClear();
		sniffer_I2CSetIntLevel();
	}
}

void sniffer_PowerOnPeripherials(void) {
	PR.PRPC = PR_TWI_bm;
	PR.PRPE = PR_TWI_bm | PR_USART0_bm | PR_USART1_bm;
}

void sniffer_PowerOffPeripherials(void) {
	PR.PRPC = PR_TWI_bm | PR_USART0_bm | PR_USART1_bm | PR_SPI_bm;
	PR.PRPE = PR_TWI_bm | PR_USART0_bm | PR_USART1_bm | PR_SPI_bm;
}

void sniffer_SPIInitPort(PORT_t *port, uint8_t option) {
	if (testbit(option, SNIFFER_SPI_CPOL)) {
		input_LogicPinConfig(port, SPI_SCK_PIN, true, INPUT_PORT_SENSE_EDGE_Rising);
	} else {
		input_LogicPinConfig(port, SPI_SCK_PIN, false, INPUT_PORT_SENSE_EDGE_Rising);
	}
	if (testbit(option, SNIFFER_SPI_SSINV)) {
		input_LogicPinConfig(port, SPI_SS_PIN, true, INPUT_PORT_SENSE_EDGE_Rising);
	} else {
		input_LogicPinConfig(port, SPI_SS_PIN, false, INPUT_PORT_SENSE_EDGE_Rising);
	}
}

void sniffer_SPIInit(uint8_t option) {
	 sniffer_SPIInitPort(&PORTC, option);
	 sniffer_SPIInitPort(&PORTE, option);
}

void sniffer_SPIStartPeripheral(SPI_t *spi, uint8_t option) {
	if (testbit(option, SNIFFER_SPI_CPHA)) {
		spi->CTRL = SPI_MODE_1_gc;
	}
	spi->CTRL |= SPI_ENABLE_bm;
}

void sniffer_SPIStart(uint8_t option) {
	sniffer_SPIStartPeripheral(&SPIC, option);
	sniffer_SPIStartPeripheral(&SPIE, option);
}

void sniffer_PortIntFlagsClear(void) {
	PORTC.INTFLAGS = PORT_INT0IF_bm | PORT_INT1IF_bm;
}

void sniffer_I2CSetIntLevel(void) {
    PORTC.INTCTRL = PORT_INT0LVL_HI_gc;
}

void sniffer_I2CInit(sniffbuffer_t buffer) {
	input_LogicPinConfig(&PORTC, I2C_SDA_PIN, false, INPUT_PORT_SENSE_EDGE_Falling);
	input_LogicPinConfig(&PORTC, I2C_SCL_PIN, false, INPUT_PORT_SENSE_EDGE_Rising);
    PORTC.INT0MASK = (1 << I2C_SDA_PIN);
	PORTC.INT1MASK = (1 << I2C_SCL_PIN);
    sniffer_PortIntFlagsClear();
	sniffer_I2CSetIntLevel();
}

uint8_t sniffer_USARTFlushBuffer(USART_t *port) {
	uint8_t dummy = 0;
	//23.7.5 Flushing the Receive Buffer
	while (testbit(port->STATUS, USART_RXCIF_bp)) {
		dummy = port->DATA;
	}
	return dummy;
}

void sniffer_USARTInit(USART_t *port, uint8_t baud, uint8_t databits, uint8_t option) {
    port->BAUDCTRLA = pgm_read_byte_near(BAUDA + baud);
    port->BAUDCTRLB = pgm_read_byte_near(BAUDB + baud);
    port->CTRLC = databits;
    port->CTRLB = USART_RXEN_bm;
	if (testbit(option, SNIFFER_RxTx_Synchronization)) {
		port->CTRLA = USART_RXCINTLVL_HI_gc;
	} else {
		port->CTRLA = 0;
	}
}

uint8_t sniffer_USARTDataLengthToUSARTRegister(uint8_t length) {
	switch (length) {
		case USART_DATA_LENGTH_5bit:
			return USART_CHSIZE_5BIT_gc;
		case USART_DATA_LENGTH_6bit:
			return USART_CHSIZE_6BIT_gc;
		case USART_DATA_LENGTH_7bit:
			return USART_CHSIZE_7BIT_gc;
		case USART_DATA_LENGTH_8bit:
			return USART_CHSIZE_8BIT_gc;
		case USART_DATA_LENGTH_9bit:
			return USART_CHSIZE_9BIT_gc;
		default:
			return USART_CHSIZE_8BIT_gc;
	}	
}

void sniffer_USARTInitProtocol(sniffconfig_t *config, uint8_t bitrate, uint8_t option) {
	uint8_t baud, databits;
	
    baud = bitrate & SNIFFER_UART_BAUTRATE_bm;
    databits = sniffer_USARTDataLengthToUSARTRegister((bitrate & SNIFFER_UART_DATABITS_bm) >> SNIFFER_UART_DATABITS_bp);
    if (testbit(option, SNIFFER_UART_ParityEnable)) {
        if (testbit(option, SNIFFER_UART_ParityOdd)) {
				databits |= USART_PMODE_ODD_gc;
	        } else {
				databits |= USART_PMODE_EVEN_gc;
			}
        }
    if (testbit(option, SNIFFER_UART_StopBit2)) databits |= USART_SBMODE_bm;
	sniffer_USARTInit(&USARTC0, baud, databits, option);
	sniffer_USARTInit(&USARTC1, baud, databits, option);
}

void sniffer_ParallelInit(uint8_t option, uint8_t edge) {
	input_LogicPortConfig(&PORTC, testbit(option, SNIFFER_PARALLEL_InputInvert), INPUT_PORT_SENSE_EDGE_Rising);
	if (edge == SNIFFER_PARALLEL_EDGE_Rising) {
		input_LogicPinConfig(&PORTD, INPUT_EXT_PIN, false, INPUT_PORT_SENSE_EDGE_Rising);
	} else if (edge == SNIFFER_PARALLEL_EDGE_Falling) {
		input_LogicPinConfig(&PORTD, INPUT_EXT_PIN, false, INPUT_PORT_SENSE_EDGE_Falling);
	} else if (edge == SNIFFER_PARALLEL_EDGE_Bothedges) {
		input_LogicPinConfig(&PORTD, INPUT_EXT_PIN, false, INPUT_PORT_SENSE_EDGE_Bothedges);
	}
}	

void sniffer_ParallelStart(void) {
	EVSYS.CH2MUX = EVSYS_CHMUX_PORTD_PIN1_gc;
}

void sniffer_InitTimeoutTimer(void) {
	TCF0.CTRLA = TC_CLKSEL_OFF_gc;
	TCF0.CTRLFSET = TC_CMD_RESET_gc;
	TCF0.CTRLE = TC_BYTEM_NORMAL_gc;
	TCF0.CNT = 0;
	TCF0.CTRLA = TC_CLKSEL_DIV1024_gc;
	TCF0.PER = 625;			// timeout 20ms
}

uint8_t sniffer_Decode(uint8_t *mainmode, sniffbuffer_t buffer, config_t *cfg, uint8_t option, uint8_t sniff) {
	uint8_t result = 0;
	bool newdata;
	
    sniffer_PowerOnPeripherials();
	input_LogicPortConfig(&PORTC, false, INPUT_PORT_SENSE_EDGE_Rising);
	input_LogicPortConfig(&PORTE, false, INPUT_PORT_SENSE_EDGE_Rising);
	data_BufferClear8(buffer.all, data_SnifferDataLength(), 0);
    if (cfg->sniffprotocol == SNIFFER_PROTOCOL_I2C) {
		EVSYS.CH2MUX = 0x00;
		sniffer_InitTimeoutTimer();
		sniffer_I2CInit(buffer);
		sniffer_DMAStart(cfg->sniffprotocol, buffer, option, sniff);
    } else if (cfg->sniffprotocol == SNIFFER_PROTOCOL_USART) {
		EVSYS.CH2MUX = 0x00;
		sniffer_USARTInitProtocol(buffer.config, cfg->bitrate, option);
		sniffer_USARTFlushBuffer(&USARTC0);
		sniffer_USARTFlushBuffer(&USARTC1);
		sniffer_DMAStart(cfg->sniffprotocol, buffer, option, sniff);
    } else if (cfg->sniffprotocol == SNIFFER_PROTOCOL_SPI) {
		sniffer_SPIInit(option);
		sniffer_DMAStart(cfg->sniffprotocol, buffer, option, sniff);
		sniffer_SPIStart(option);
    } else if (cfg->sniffprotocol == SNIFFER_PROTOCOL_Parallel) {
		sniffer_ParallelInit(option, cfg->sniffportedge);
		sniffer_DMAStart(cfg->sniffprotocol, buffer, option, sniff);
		sniffer_ParallelStart();
	}
	newdata = true;
	buffer.config->id = 0;
    while ((*mainmode == MAIN_MODE_Sniff) && (testbit(GPIO_Action, SNIFFER_REQUEST_Run))) {
		if (testbit(GPIO_Action, SNIFFER_REQUEST_Clear)) {
			clrbit(GPIO_Action, SNIFFER_REQUEST_Clear);
			sniffer_ClearBuffer(cfg->sniffprotocol, buffer, option, sniff);
			newdata = true;
		}
        if (buffer.config->indexrx != (sniffer_DMALength(cfg->sniffprotocol) - DMA.CH0.TRFCNT)) {
			buffer.config->indexrx = sniffer_DMALength(cfg->sniffprotocol) - DMA.CH0.TRFCNT;
			newdata = true;
		}
        if (buffer.config->indextx != (sniffer_DMALength(cfg->sniffprotocol) - DMA.CH1.TRFCNT)) {
			buffer.config->indextx = sniffer_DMALength(cfg->sniffprotocol) - DMA.CH1.TRFCNT;
			newdata = true;
		}
        if (testbit(sniff, MAIN_SNIFFER_SingleSniff)) {
            if (!testbit(DMA.CH0.CTRLA, DMA_CH_ENABLE_bp)) {
				clrbit(GPIO_Action, SNIFFER_REQUEST_Run);
				PORTC.INTCTRL = 0x00;
				buffer.config->indexrx = sniffer_DMALength(cfg->sniffprotocol) - 1;
            }
        }
        if (testbit(sniff, MAIN_SNIFFER_SingleSniff) && (cfg->sniffprotocol != SNIFFER_PROTOCOL_Parallel)) {
	        if (!testbit(DMA.CH1.CTRLA, DMA_CH_ENABLE_bp)) {
	            clrbit(GPIO_Action, SNIFFER_REQUEST_Run);
				PORTC.INTCTRL = 0x00;
				buffer.config->indextx = sniffer_DMALength(cfg->sniffprotocol) - 1;
	        }
        }
		if (newdata) {
			newdata = false;
			buffer.config->type = FRAME_HEADER_NAME_Sniffer;
			buffer.config->protocol = cfg->sniffprotocol;
			buffer.config->id += 1;
			buffer.config->bitrate = cfg->bitrate;
			buffer.config->option = option;
			result = usb_SendDataBuffer(buffer.all, data_SnifferDataLength());
			if (result) {
				clrbit(GPIO_Action, SNIFFER_REQUEST_Run);
				sniffer_Stop();
				data_BufferClear8(buffer.all, data_SnifferDataLength(), 0);
				return result;
			}
		}
    } 
	sniffer_Stop();
	data_BufferClear8(buffer.all, data_SnifferDataLength(), 0);
	return result;
}

void sniffer_USARTStopPort(USART_t *port) {
   	port->CTRLA = 0x00;
   	port->CTRLB = 0x00;
}

void sniffer_USARTStop(void) {
	sniffer_USARTStopPort(&USARTC0);
	sniffer_USARTStopPort(&USARTC1);	
}

void sniffer_SPIStopPort(SPI_t *spi) {
	SPIC.INTCTRL = 0x00;
   	SPIC.CTRL = 0x00;
}

void sniffer_SPIStop(void) {
	sniffer_SPIStopPort(&SPIC);
	sniffer_SPIStopPort(&SPIE);
}

void sniffer_I2CStop(void) {
   	PORTC.INTCTRL = 0x00;
    PORTC.INT0MASK = 0x00;
	PORTC.INT1MASK = 0x00;
}

void sniffer_Stop(void) {
	sniffer_I2CStop();
   	sniffer_SPIStop();
	sniffer_DMAStop();
	sniffer_USARTStop();
	EVSYS.CH2MUX = 0x00;
	input_LogicPortConfig(&PORTC, testbit(GPIO_CHDctrl, MAIN_CHANNEL_CONTROL_Invert), INPUT_PORT_SENSE_EDGE_Rising);
	input_LogicPinConfig(&PORTD, INPUT_EXT_PIN, false, INPUT_PORT_SENSE_EDGE_Rising);
   	sniffer_PowerOffPeripherials();
}

// TWI look for start condition
ISR(PORTC_INT0_vect) {  
	if (!SCL) return;
    // wait for SCL to fall
    TCF0.CTRLFSET = TC_CMD_RESTART_gc;
    while(SCL) if (testbit(TCF0.INTFLAGS, TC0_OVFIF_bp)) goto timeout;
	PORTC.INTFLAGS = PORT_INT0IF_bm | PORT_INT1IF_bm;
	PORTC.INTCTRL = PORT_INT1LVL_HI_gc;
	txdata = SNIFFER_I2C_Start | SNIFFER_I2C_ValidData;
	EVSYS.STROBE = EVSYS_CH2_bm;
	i2cstart = true;
	i2cfirst = true;
	return;
timeout:
	txdata = SNIFFER_I2C_Timeout | SNIFFER_I2C_ValidData;
	EVSYS.STROBE = EVSYS_CH2_bm;
	PORTC.INTFLAGS = PORT_INT0IF_bm;
}

// TWI look for data bytes
// Base on i2c protocol sniffer written by Bill Grundmann
ISR(PORTC_INT1_vect) {
	uint8_t dt, i2cbits = 0;
	
	if (!i2cstart) return;
sample_bit:    
	// now ready for next bit of the byte
	TCF0.CTRLFSET = TC_CMD_RESTART_gc;
	while(SCL == 0) if (testbit(TCF0.INTFLAGS, TC0_OVFIF_bp)) goto timeout;  // wait for rising edge
	dt = SDA;   
	// Sampling SDA on the rising edge of SCL
	// we might not use this data in the cases of either stop or repeated start conditions
	// SCL is now high, wait for either of two conditions: SCL falls or SDA changes
	// split the cases to reduce the sub-loop overheads
	if (dt == 0) {
		// special case where stop condition might exist because data was zero
		TCF0.CTRLFSET = TC_CMD_RESTART_gc;
look_for_bit_or_stop:
		if (testbit(TCF0.INTFLAGS, TC0_OVFIF_bp)) goto timeout; 
		if (SCL == 0) goto process_bit;
		if (SDA == 0) goto look_for_bit_or_stop;
		// double check, in case both signals changed at the same time
		if (SCL == 0) goto process_bit; 
		// SCL and SDA are now high (stop condition occurred)
		PORTC.INTFLAGS = PORT_INT0IF_bm | PORT_INT1IF_bm;
		PORTC.INTCTRL = PORT_INT0LVL_HI_gc;
		txdata = SNIFFER_I2C_Stop | SNIFFER_I2C_ValidData;
		EVSYS.STROBE = EVSYS_CH2_bm;
		i2cstart = false;
		return;
	} else {
		// special case where we might have a repeated start where SDA falls while SCL high
		TCF0.CTRLFSET = TC_CMD_RESTART_gc;
look_for_bit_or_repeated_start:
		if (testbit(TCF0.INTFLAGS, TC0_OVFIF_bp)) goto timeout;
		if (SCL == 0) goto process_bit;
		if (SDA) goto look_for_bit_or_repeated_start;
		// double check, in case both signals changed at the same time
		if (SCL == 0) goto process_bit; 
		// SCL is high and SDA is now low (repeated start condition occurred)
		txdata = SNIFFER_I2C_RepeatedStart | SNIFFER_I2C_ValidData;
		EVSYS.STROBE = EVSYS_CH2_bm;
		PORTC.INTFLAGS = PORT_INT1IF_bm;
		i2cfirst = true;
		return;
	}
process_bit:
	// process another data bit
	if (i2cbits == 8) {
		// byte complete
		txdata = SNIFFER_I2C_ValidData;
		if (i2cfirst) txdata |= SNIFFER_I2C_Address;
		if (dt == 0) txdata |= SNIFFER_I2C_ACK;
		EVSYS.STROBE = EVSYS_CH2_bm;
		PORTC.INTFLAGS = PORT_INT1IF_bm;
		i2cfirst = false;
		return;
	} else {
		// accumulate bits of the byte's information (MSB to LSB)
		rxdata = (rxdata << 1) | dt;
		i2cbits++;
	}
	goto sample_bit;
timeout:
	txdata = SNIFFER_I2C_ValidData;
	if (SDA && SCL && (i2cbits == 0)) {
		txdata |= SNIFFER_I2C_FastStop;
	} else {
		txdata |= SNIFFER_I2C_Timeout;
	}
	EVSYS.STROBE = EVSYS_CH2_bm;
	PORTC.INTFLAGS = PORT_INT0IF_bm | PORT_INT1IF_bm;
	PORTC.INTCTRL = PORT_INT0LVL_HI_gc;
}

// rxdata holds data USARTC0
ISR(USARTC0_RXC_vect) {
	rxdata = USARTC0.DATA;
	txdata = 0;
	dxdata = SNIFFER_VALID_Rx_DATA_Bit;
	EVSYS.STROBE = EVSYS_CH2_bm;
}

// txdata holds data USARTC1
ISR(USARTC1_RXC_vect) {
	rxdata = 0;
	txdata = USARTC1.DATA;
	dxdata = SNIFFER_VALID_Tx_DATA_Bit;
	EVSYS.STROBE = EVSYS_CH2_bm;
}
